import asyncio
import logging
import time
import warnings
from datetime import datetime
from socket import timeout
from typing import KeysView

import numpy as np
from cbpi.api import *
from cbpi.api import Property, action, parameters
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType
from cbpi.api.dataclasses import Kettle, NotificationAction, NotificationType, Props
from cbpi.api.step import CBPiStep, StepResult
from cbpi.api.timer import Timer
from voluptuous.schema_builder import message


@parameters(
    [
        Property.Text(
            label="Notification", configurable=True, description="Text for notification"
        ),
        Property.Select(
            label="AutoNext",
            options=["Yes", "No"],
            description="Automatically move to next step (Yes) or pause after Notification (No)",
        ),
    ]
)
class NotificationStep(CBPiStep):

    async def NextStep(self, **kwargs):
        await self.next()

    async def on_timer_done(self, timer):
        self.summary = self.props.get("Notification", "")

        if self.AutoNext == True:
            self.cbpi.notify(
                self.name, self.props.get("Notification", ""), NotificationType.INFO
            )
            await self.next()
        else:
            self.cbpi.notify(
                self.name,
                self.props.get("Notification", ""),
                NotificationType.INFO,
                action=[NotificationAction("Next Step", self.NextStep)],
            )
            await self.push_update()

    async def on_timer_update(self, timer, seconds):
        await self.push_update()

    async def on_start(self):
        self.summary = ""
        self.AutoNext = False if self.props.get("AutoNext", "No") == "No" else True
        if self.timer is None:
            self.timer = Timer(
                1, on_update=self.on_timer_update, on_done=self.on_timer_done
            )
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
            if self.timer.is_running is not True:
                self.timer.start()
                self.timer.is_running = True

        return StepResult.DONE


@parameters(
    [
        Property.Number(label="Temp", configurable=True),
        Property.Sensor(label="Sensor"),
        Property.Kettle(label="Kettle"),
        Property.Text(
            label="Notification",
            configurable=True,
            description="Text for notification when Temp is reached",
        ),
        Property.Select(
            label="AutoMode",
            options=["Yes", "No"],
            description="Switch Kettlelogic automatically on and off -> Yes",
        ),
    ]
)
class MashInStep(CBPiStep):

    async def NextStep(self, **kwargs):
        await self.next()

    async def on_timer_done(self, timer):
        self.summary = ""
        self.kettle.target_temp = 0
        await self.push_update()
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(
            self.name,
            self.props.get(
                "Notification",
                "Target Temp reached. Please add malt and klick next to move on.",
            ),
            action=[NotificationAction("Next Step", self.NextStep)],
        )

    async def on_timer_update(self, timer, seconds):
        await self.push_update()

    async def on_start(self):
        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        self.kettle = self.get_kettle(self.props.get("Kettle", None))
        if self.kettle is not None:
            self.kettle.target_temp = int(self.props.get("Temp", 0))
        if self.AutoMode == True:
            await self.setAutoMode(True)
        self.summary = "Waiting for Target Temp"
        if self.cbpi.kettle is not None and self.timer is None:
            self.timer = Timer(
                1, on_update=self.on_timer_update, on_done=self.on_timer_done
            )
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
            sensor_value = self.get_sensor_value(self.props.get("Sensor", None)).get(
                "value"
            )
            if (
                sensor_value >= int(self.props.get("Temp", 0))
                and self.timer.is_running is not True
            ):
                self.timer.start()
                self.timer.is_running = True
        await self.push_update()
        return StepResult.DONE

    async def reset(self):
        self.timer = Timer(
            1, on_update=self.on_timer_update, on_done=self.on_timer_done
        )

    async def setAutoMode(self, auto_state):
        try:
            if (
                self.kettle.instance is None or self.kettle.instance.state == False
            ) and (auto_state is True):
                await self.cbpi.kettle.toggle(self.kettle.id)
            elif (self.kettle.instance.state == True) and (auto_state is False):
                await self.cbpi.kettle.stop(self.kettle.id)
            await self.push_update()

        except Exception as e:
            logging.error(
                "Failed to switch on KettleLogic {} {}".format(self.kettle.id, e)
            )


@parameters(
    [
        Property.Number(
            label="Timer", description="Time in Minutes", configurable=True
        ),
        Property.Number(label="Temp", configurable=True),
        Property.Sensor(label="Sensor"),
        Property.Kettle(label="Kettle"),
        Property.Select(
            label="AutoMode",
            options=["Yes", "No"],
            description="Switch Kettlelogic automatically on and off -> Yes",
        ),
    ]
)
class MashStep(CBPiStep):

    @action("Start Timer", [])
    async def start_timer(self):
        if self.timer.is_running is not True:
            self.cbpi.notify(self.name, "Timer started", NotificationType.INFO)
            self.timer.start()
            self.timer.is_running = True
        else:
            self.cbpi.notify(
                self.name, "Timer is already running", NotificationType.WARNING
            )

    @action("Add 5 Minutes to Timer", [])
    async def add_timer(self):
        if self.timer.is_running == True:
            self.cbpi.notify(self.name, "5 Minutes added", NotificationType.INFO)
            await self.timer.add(300)
        else:
            self.cbpi.notify(
                self.name, "Timer must be running to add time", NotificationType.WARNING
            )

    async def on_timer_done(self, timer):
        self.summary = ""
        self.kettle.target_temp = 0
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(self.name, "Step finished", NotificationType.SUCCESS)

        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        self.kettle = self.get_kettle(self.props.Kettle)
        if self.kettle is not None:
            self.kettle.target_temp = int(self.props.get("Temp", 0))
        if self.AutoMode == True:
            await self.setAutoMode(True)
        await self.push_update()

        if self.cbpi.kettle is not None and self.timer is None:
            self.timer = Timer(
                int(self.props.get("Timer", 0)) * 60,
                on_update=self.on_timer_update,
                on_done=self.on_timer_done,
            )
        elif self.cbpi.kettle is not None:
            try:
                if self.timer.is_running == True:
                    self.timer.start()
            except:
                pass

        self.summary = "Waiting for Target Temp"
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def reset(self):
        self.timer = Timer(
            int(self.props.get("Timer", 0)) * 60,
            on_update=self.on_timer_update,
            on_done=self.on_timer_done,
        )

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
            sensor_value = self.get_sensor_value(self.props.get("Sensor", None)).get(
                "value"
            )
            if (
                sensor_value >= int(self.props.get("Temp", 0))
                and self.timer.is_running is not True
            ):
                self.timer.start()
                self.timer.is_running = True
                estimated_completion_time = datetime.fromtimestamp(
                    time.time() + (int(self.props.get("Timer", 0))) * 60
                )
                self.cbpi.notify(
                    self.name,
                    "Timer started. Estimated completion: {}".format(
                        estimated_completion_time.strftime("%H:%M")
                    ),
                    NotificationType.INFO,
                )
        return StepResult.DONE

    async def setAutoMode(self, auto_state):
        try:
            if (
                self.kettle.instance is None or self.kettle.instance.state == False
            ) and (auto_state is True):
                await self.cbpi.kettle.toggle(self.kettle.id)
            elif (self.kettle.instance.state == True) and (auto_state is False):
                await self.cbpi.kettle.stop(self.kettle.id)
            await self.push_update()

        except Exception as e:
            logging.error(
                "Failed to switch on KettleLogic {} {}".format(self.kettle.id, e)
            )


@parameters(
    [Property.Number(label="Timer", description="Time in Minutes", configurable=True)]
)
class WaitStep(CBPiStep):

    async def on_timer_done(self, timer):
        self.summary = ""
        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        if self.timer is None:
            self.timer = Timer(
                int(self.props.Timer) * 60,
                on_update=self.on_timer_update,
                on_done=self.on_timer_done,
            )
        self.timer.start()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def reset(self):
        self.timer = Timer(
            int(self.props.Timer) * 60,
            on_update=self.on_timer_update,
            on_done=self.on_timer_done,
        )

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
        return StepResult.DONE


@parameters(
    [
        Property.Select(
            label="toggle_type",
            options=["On", "Off"],
            description="Choose if Actor should be switched on or off in this step",
        ),
        Property.Actor(
            label="Actor", description="Actor that should be toggled during this step"
        ),
    ]
)
class ToggleStep(CBPiStep):
    async def on_timer_done(self, timer):
        self.summary = ""
        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        if self.timer is None:
            self.timer = Timer(
                1, on_update=self.on_timer_update, on_done=self.on_timer_done
            )
        self.timer.start()
        self.type = self.props.get("toggle_type", "Off")
        self.Actor = self.props.get("Actor", None)
        if self.Actor is not None and self.type == "On":
            await self.actor_on(self.Actor)
        if self.Actor is not None and self.type == "Off":
            await self.actor_off(self.Actor)

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def reset(self):
        self.timer = Timer(
            1, on_update=self.on_timer_update, on_done=self.on_timer_done
        )

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
        return StepResult.DONE


@parameters(
    [
        Property.Number(
            label="Timer", description="Time in Minutes", configurable=True
        ),
        Property.Actor(label="Actor"),
    ]
)
class ActorStep(CBPiStep):
    async def on_timer_done(self, timer):
        self.summary = ""
        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        if self.timer is None:
            self.timer = Timer(
                int(self.props.Timer) * 60,
                on_update=self.on_timer_update,
                on_done=self.on_timer_done,
            )
        self.timer.start()
        await self.actor_on(self.props.Actor)

    async def on_stop(self):
        await self.actor_off(self.props.Actor)
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def reset(self):
        self.timer = Timer(
            int(self.props.Timer) * 60,
            on_update=self.on_timer_update,
            on_done=self.on_timer_done,
        )

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
        return StepResult.DONE


@parameters(
    [
        Property.Number(
            label="Timer", description="Time in Minutes", configurable=True
        ),
        Property.Number(
            label="Temp", description="Boil temperature", configurable=True
        ),
        # Property.Number(label="DwellTime", description="Time in minutes (Recommended: 5). If temp is not changing significantly within this time, Timer will start. Defaul:0 disabled", configurable=True),
        Property.Sensor(label="Sensor"),
        Property.Kettle(label="Kettle"),
        Property.Select(
            label="LidAlert",
            options=["Yes", "No"],
            description="Trigger Alert to remove lid if temp is close to boil",
        ),
        Property.Select(
            label="AutoMode",
            options=["Yes", "No"],
            description="Switch Kettlelogic automatically on and off -> Yes",
        ),
        # Property.Select(label="AutoTimer",options=["Yes","No"], description="Start Timer automatically if Temp does not change for 5 Minutes and is above 95C/203F"),
        Property.Select(
            "First_Wort",
            options=["Yes", "No"],
            description="First Wort Hop alert if set to Yes",
        ),
        Property.Text(
            "First_Wort_text",
            configurable=True,
            description="First Wort Hop alert text",
        ),
        Property.Number(
            "Hop_1",
            configurable=True,
            description="First Hop alert (minutes before finish)",
        ),
        Property.Text(
            "Hop_1_text", configurable=True, description="First Hop alert text"
        ),
        Property.Number(
            "Hop_2",
            configurable=True,
            description="Second Hop alert (minutes before finish)",
        ),
        Property.Text(
            "Hop_2_text", configurable=True, description="Second Hop alert text"
        ),
        Property.Number(
            "Hop_3",
            configurable=True,
            description="Third Hop alert (minutes before finish)",
        ),
        Property.Text(
            "Hop_3_text", configurable=True, description="Third Hop alert text"
        ),
        Property.Number(
            "Hop_4",
            configurable=True,
            description="Fourth Hop alert (minutes before finish)",
        ),
        Property.Text(
            "Hop_4_text", configurable=True, description="Fourth Hop alert text"
        ),
        Property.Number(
            "Hop_5",
            configurable=True,
            description="Fifth Hop alert (minutes before finish)",
        ),
        Property.Text(
            "Hop_5_text", configurable=True, description="Fifth Hop alert text"
        ),
        Property.Number(
            "Hop_6",
            configurable=True,
            description="Sixth Hop alert (minutes before finish)",
        ),
        Property.Text(
            "Hop_6_text", configurable=True, description="Sixth Hop alert text"
        ),
    ]
)
class BoilStep(CBPiStep):

    @action("Start Timer", [])
    async def start_timer(self):
        if self.timer.is_running is not True:
            self.cbpi.notify(self.name, "Timer started", NotificationType.INFO)
            self.timer.start()
            self.timer.is_running = True
        else:
            self.cbpi.notify(
                self.name, "Timer is already running", NotificationType.WARNING
            )

    @action("Add 5 Minutes to Timer", [])
    async def add_timer(self):
        if self.timer.is_running == True:
            self.cbpi.notify(self.name, "5 Minutes added", NotificationType.INFO)
            await self.timer.add(300)
        else:
            self.cbpi.notify(
                self.name, "Timer must be running to add time", NotificationType.WARNING
            )

    async def on_timer_done(self, timer):
        self.summary = ""
        self.kettle.target_temp = 0
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(self.name, "Boiling completed", NotificationType.SUCCESS)
        await self.next()

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        self.remaining_seconds = seconds
        await self.push_update()

    async def on_start(self):

        self.lid_temp = 95 if self.get_config_value("TEMP_UNIT", "C") == "C" else 203
        self.lid_flag = True if self.props.get("LidAlert", "No") == "Yes" else False
        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        self.AutoTimer = (
            True if self.cbpi.config.get("BoilAutoTimer", "No") == "Yes" else False
        )
        self.first_wort_hop_flag = False
        self.first_wort_hop = self.props.get("First_Wort", "No")
        self.first_wort_hop_text = self.props.get("First_Wort_text", None)
        self.hops_added = ["", "", "", "", "", ""]
        self.remaining_seconds = None
        self.temparray = np.array([])
        # self.dwelltime=int(self.props.get("DwellTime", 0))*60
        self.dwelltime = (
            5 * 60
        )  # tested with 5 minutes -> not exactly 5 min due to accuracy of asyncio.sleep
        self.deviationlimit = 0.3  # derived from a test
        # logging.warning(self.AutoTimer)
        self.summary2 = None

        self.kettle = self.get_kettle(self.props.get("Kettle", None))
        if self.kettle is not None:
            self.kettle.target_temp = float(self.props.get("Temp", 0))

        if self.cbpi.kettle is not None and self.timer is None:
            self.timer = Timer(
                int(self.props.get("Timer", 0)) * 60,
                on_update=self.on_timer_update,
                on_done=self.on_timer_done,
            )

        elif self.cbpi.kettle is not None:
            try:
                if self.timer.is_running == True:
                    self.timer.start()
            except:
                pass

        self.summary = "Waiting for Target Temp"
        if self.AutoMode == True:
            await self.setAutoMode(True)
        await self.push_update()

    async def next_hop_timer(self):
        hop_timers = []
        for x in range(1, 6):
            try:
                hop = int(self.props.get("Hop_%s" % x, None)) * 60
            except:
                hop = None
            if hop is not None and self.remaining_seconds is not None:
                hop_left = self.remaining_seconds - hop
                if hop_left > 0:
                    hop_timers.append(hop_left)
        if hop_timers:
            next_hop_timer = time.strftime("%H:%M:%S", time.gmtime(min(hop_timers)))
        else:
            next_hop_timer = None
        return next_hop_timer

    async def check_hop_timer(self, number, value, text):
        if value is not None and self.hops_added[number - 1] is not True:
            if self.remaining_seconds != None and self.remaining_seconds <= (
                int(value) * 60 + 1
            ):
                self.hops_added[number - 1] = True
                if text is not None and text != "":
                    self.cbpi.notify(
                        "Hop Alert",
                        "Please add %s (%s)" % (text, number),
                        NotificationType.INFO,
                    )
                else:
                    self.cbpi.notify(
                        "Hop Alert", "Please add Hop %s" % number, NotificationType.INFO
                    )

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        self.summary2 = None
        self.kettle.target_temp = 0
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def reset(self):
        self.timer = Timer(
            int(self.props.get("Timer", 0)) * 60,
            on_update=self.on_timer_update,
            on_done=self.on_timer_done,
        )

    async def run(self):
        if self.first_wort_hop_flag == False and self.first_wort_hop == "Yes":
            self.first_wort_hop_flag = True
            if self.first_wort_hop_text is not None and self.first_wort_hop_text != "":
                self.cbpi.notify(
                    "First Wort Hop Addition!",
                    "Please add %s for first wort" % self.first_wort_hop_text,
                    NotificationType.INFO,
                )
            else:
                self.cbpi.notify(
                    "First Wort Hop Addition!",
                    "Please add hops for first wort",
                    NotificationType.INFO,
                )

        while self.running == True:
            await asyncio.sleep(1)
            sensor_value = self.get_sensor_value(self.props.get("Sensor", None)).get(
                "value"
            )
            self.temparray = np.append(self.temparray, sensor_value)
            if self.temparray.size > self.dwelltime:
                self.temparray = np.delete(self.temparray, 0)
                deviation = np.std(self.temparray)
                if (
                    (sensor_value >= self.lid_temp)
                    and (deviation <= self.deviationlimit)
                    and (self.AutoTimer is True)
                    and (self.timer.is_running is not True)
                ):
                    self.timer.start()
                    self.timer.is_running = True
                    estimated_completion_time = datetime.fromtimestamp(
                        time.time() + (int(self.props.get("Timer", 0))) * 60
                    )
                    self.cbpi.notify(
                        self.name,
                        "Timer started automatically. Estimated completion: {}".format(
                            estimated_completion_time.strftime("%H:%M")
                        ),
                        NotificationType.INFO,
                    )
                logging.info(
                    "Current: " + str(sensor_value) + " | Dev: " + str(deviation)
                )
            if self.lid_flag == True and sensor_value >= self.lid_temp:
                self.cbpi.notify(
                    "Please remove lid!",
                    "Reached temp close to boiling",
                    NotificationType.INFO,
                )
                self.lid_flag = False

            if (
                sensor_value >= float(self.props.get("Temp", 0))
                and self.timer.is_running is not True
            ):
                self.timer.start()
                self.timer.is_running = True
                estimated_completion_time = datetime.fromtimestamp(
                    time.time() + (int(self.props.get("Timer", 0))) * 60
                )
                self.cbpi.notify(
                    self.name,
                    "Timer started. Estimated completion: {}".format(
                        estimated_completion_time.strftime("%H:%M")
                    ),
                    NotificationType.INFO,
                )
            else:
                if self.timer.is_running:
                    try:
                        nexthoptimer = await self.next_hop_timer()
                        self.summary2 = (
                            "Add Hop in: %s" % nexthoptimer
                            if nexthoptimer is not None
                            else None
                        )
                    except:
                        self.summary2 = None
                for x in range(1, 6):
                    await self.check_hop_timer(
                        x,
                        self.props.get("Hop_%s" % x, None),
                        self.props.get("Hop_%s_text" % x, None),
                    )

        return StepResult.DONE

    async def setAutoMode(self, auto_state):
        try:
            if (
                self.kettle.instance is None or self.kettle.instance.state == False
            ) and (auto_state is True):
                await self.cbpi.kettle.toggle(self.kettle.id)
            elif (self.kettle.instance.state == True) and (auto_state is False):
                await self.cbpi.kettle.stop(self.kettle.id)
            await self.push_update()

        except Exception as e:
            logging.error(
                "Failed to switch on KettleLogic {} {}".format(self.kettle.id, e)
            )


@parameters(
    [
        Property.Number(
            label="Temp",
            configurable=True,
            description="Target temperature for cooldown. Notification will be send when temp is reached and Actor can be triggered",
        ),
        Property.Sensor(
            label="Sensor", description="Sensor that is used during cooldown"
        ),
        Property.Actor(
            label="Actor",
            description="Actor can trigger a valve for the cooldwon to target temperature",
        ),
        Property.Kettle(label="Kettle"),
    ]
)
class CooldownStep(CBPiStep):

    async def on_timer_done(self, timer):
        self.summary = ""
        if self.actor is not None:
            await self.actor_off(self.actor)
        self.cbpi.notify(
            "CoolDown",
            "Wort cooled down. Please transfer to Fermenter.",
            NotificationType.INFO,
        )
        await self.next()

    async def on_timer_update(self, timer, seconds):
        await self.push_update()

    async def on_start(self):
        try:
            warnings.simplefilter("ignore", np.exceptions.RankWarning)
        except Exception as e:
            logging.error(f"Numpy Error: {e}")
        self.temp_array = []
        self.time_array = []
        self.kettle = self.get_kettle(self.props.get("Kettle", None))
        self.actor = self.props.get("Actor", None)
        self.target_temp = int(self.props.get("Temp", 0))
        self.Interval = (
            10  # Interval in minutes on how often cooldwon end time is calculated
        )

        self.cbpi.notify(
            self.name,
            "Cool down to {}°".format(self.target_temp),
            NotificationType.INFO,
        )
        if self.timer is None:
            self.timer = Timer(
                1, on_update=self.on_timer_update, on_done=self.on_timer_done
            )
        self.start_time = time.time()
        self.temp_array.append(
            self.get_sensor_value(self.props.get("Sensor", None)).get("value")
        )
        self.time_array.append(time.time())
        self.next_check = self.start_time + self.Interval * 60
        self.count = 0
        self.initial_date = None

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.actor is not None:
            await self.actor_off(self.actor)
        await self.push_update()

    async def reset(self):
        self.timer = Timer(
            1, on_update=self.on_timer_update, on_done=self.on_timer_done
        )

    async def run(self):
        timestring = datetime.fromtimestamp(self.start_time)
        if self.actor is not None:
            await self.actor_on(self.actor)
        self.summary = "Started: {}".format(timestring.strftime("%H:%M"))
        await self.push_update()
        while self.running == True:
            current_temp = self.get_sensor_value(self.props.get("Sensor", None)).get(
                "value"
            )
            if self.count == 3:
                self.temp_array.append(current_temp)
                current_time = time.time()
                if self.initial_date == None:
                    self.initial_date = current_time
                self.time_array.append(current_time)
                self.count = 0
            if time.time() >= self.next_check:
                self.next_check = time.time() + (self.Interval * 60)

                cooldown_model = np.polynomial.polynomial.Polynomial.fit(
                    self.temp_array, self.time_array, 2
                )
                target_time = cooldown_model(self.target_temp)
                target_timestring = datetime.fromtimestamp(target_time)
                self.summary = "ECT: {}".format(target_timestring.strftime("%H:%M"))
                self.cbpi.notify(
                    "Cooldown Step",
                    "Current: {}°, reaching {}° at {}".format(
                        round(current_temp, 1),
                        self.target_temp,
                        target_timestring.strftime("%d.%m %H:%M"),
                    ),
                    NotificationType.INFO,
                )
                self.temp_array = []
                self.time_array = []
                self.temp_array.append(
                    self.get_sensor_value(self.props.get("Sensor", None)).get("value")
                )
                self.time_array.append(time.time())
                await self.push_update()

            if current_temp <= self.target_temp and self.timer.is_running is not True:
                self.timer.start()
                self.timer.is_running = True

            self.count += 1
            await asyncio.sleep(1)

        return StepResult.DONE


def setup(cbpi):
    """
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    """

    cbpi.plugin.register("MashInStep", MashInStep)
    cbpi.plugin.register("MashStep", MashStep)
    cbpi.plugin.register("BoilStep", BoilStep)
    cbpi.plugin.register("CooldownStep", CooldownStep)
    cbpi.plugin.register("WaitStep", WaitStep)
    cbpi.plugin.register("ToggleStep", ToggleStep)
    cbpi.plugin.register("ActorStep", ActorStep)
    cbpi.plugin.register("NotificationStep", NotificationStep)



from aiohttp import web
from aiohttp_auth import auth
from cbpi.api import *


class Login:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.cbpi.register(self, url_prefix="/")
        self.db = {
            cbpi.static_config.get("username", "cbpi"): cbpi.static_config.get(
                "password", "cbpi"
            )
        }

    @request_mapping(path="/logout", name="Logout", method="GET", auth_required=True)
    async def logout_view(self, request):
        await auth.forget(request)
        return web.Response(body="OK".encode("utf-8"))

    @request_mapping(path="/login", name="Login", method="POST", auth_required=False)
    async def login_view(self, request):

        params = await request.post()

        user = params.get("username", None)

        if user in self.db and params.get("password", None) == str(self.db[user]):
            # User is in our database, remember their login details
            await auth.remember(request, user)
            return web.Response(body="OK".encode("utf-8"))

        raise web.HTTPForbidden()



__all__ = [
    "CBPiActor",
    "CBPiExtension",
    "Property",
    "PropertyType",
    "on_event",
    "on_startup",
    "request_mapping",
    "action",
    "parameters",
    "background_task",
    "CBPiKettleLogic",
    "CBPiFermenterLogic",
    "CBPiException",
    "KettleException",
    "SensorException",
    "ActorException",
    "CBPiSensor",
    "CBPiStep",
    "CBPiFermentationStep",
]

from cbpi.api.actor import *
from cbpi.api.decorator import *
from cbpi.api.exceptions import *
from cbpi.api.extension import *
from cbpi.api.fermenter_logic import *
from cbpi.api.kettle_logic import *
from cbpi.api.property import *
from cbpi.api.sensor import *
from cbpi.api.step import *



import asyncio
import logging
from abc import ABCMeta, abstractmethod

from cbpi.api.base import CBPiBase
from cbpi.api.dataclasses import DataType
from cbpi.api.extension import CBPiExtension


class CBPiSensor(CBPiBase, metaclass=ABCMeta):

    def __init__(self, cbpi, id, props):
        self.cbpi = cbpi
        self.id = id
        self.props = props
        self.logger = logging.getLogger(__file__)
        self.data_logger = None
        self.state = False
        self.running = False
        self.datatype = DataType.VALUE
        self.inrange = True
        self.temprange = 0
        self.kettle = None
        self.fermenter = None

    def init(self):
        pass

    def log_data(self, value):
        self.cbpi.log.log_data(self.id, value)

    def get_state(self):
        pass

    def get_value(self):
        pass

    def get_unit(self):
        pass

    def checkrange(self, value):
        # if Kettle and fermenter are selected, range check is deactivated
        if self.kettle is not None and self.fermenter is not None:
            return True
        try:
            if self.kettle is not None:
                target_temp = float(self.kettle.target_temp)
            if self.fermenter is not None:
                target_temp = float(self.fermenter.target_temp)

            diff = abs(target_temp - value)
            if diff > self.temprange:
                return False
            else:
                return True
        except Exception as e:
            return True

    def push_update(self, value, mqtt=True):
        if self.temprange != 0:
            self.inrange = self.checkrange(value)
        else:
            self.inrange = True
        try:
            self.cbpi.ws.send(
                dict(
                    topic="sensorstate",
                    id=self.id,
                    value=value,
                    datatype=self.datatype.value,
                    inrange=self.inrange,
                )
            )
            if mqtt:
                self.cbpi.push_update(
                    "cbpi/sensordata/{}".format(self.id),
                    dict(
                        id=self.id,
                        value=value,
                        datatype=self.datatype.value,
                        inrange=self.inrange,
                    ),
                    retain=True,
                )
        #            self.cbpi.push_update("cbpi/sensor/{}/udpate".format(self.id), dict(id=self.id, value=value), retain=True)
        except:
            logging.error("Failed to push sensor update for sensor {}".format(self.id))

    async def start(self):
        pass

    async def stop(self):
        pass

    async def on_start(self):
        pass

    async def on_stop(self):
        pass

    async def run(self):
        pass

    async def _run(self):

        try:
            await self.on_start()
            self.cancel_reason = await self.run()
        except asyncio.CancelledError as e:
            pass
        finally:
            await self.on_stop()



from cbpi.api import Property, parameters

from . import GenericMqttActor


@parameters(
    [
        Property.Text(label="Topic", configurable=True, description="MQTT Topic"),
    ]
)
class TasmotaMqttActor(GenericMqttActor):
    def __init__(self, cbpi, id, props):
        GenericMqttActor.__init__(self, cbpi, id, props)

    async def on_start(self):
        await GenericMqttActor.on_start(self)
        self.payload = "{switch_onoff}"



from cbpi.utils.encoder import ComplexEncoder

__all__ = ["load_config", "json_dumps"]

import json

import yaml


def load_config(fname):

    try:
        with open(fname, "rt") as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        return data
    except Exception as e:

        pass


def json_dumps(obj):
    return json.dumps(obj, cls=ComplexEncoder)



import asyncio
import json
import logging
import webbrowser
from asyncio import tasks

import aiohttp
from aiohttp import web
from cbpi.api import *
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType
from cbpi.api.dataclasses import Fermenter, Props, Step
from cbpi.controller.fermentation_controller import FermentationController


class FermenterAutostart(CBPiExtension):

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self._task = asyncio.create_task(self.run())
        self.controller: FermentationController = cbpi.fermenter

    async def run(self):
        logging.info("Starting Fermenter Autorun")
        # get all kettles
        try:
            self.fermenter = self.controller.get_state()
            for id in self.fermenter["data"]:
                try:
                    self.autostart = id["props"]["AutoStart"]
                    if self.autostart == "Yes":
                        fermenter_id = id["id"]
                        logging.info(
                            "Enabling Autostart for Fermenter {}".format(fermenter_id)
                        )
                        self.fermenter = self.cbpi.fermenter._find_by_id(fermenter_id)
                        try:
                            if (
                                self.fermenter.instance is None
                                or self.fermenter.instance.state == False
                            ):
                                await self.cbpi.fermenter.start(self.fermenter.id)
                                logging.info(
                                    "Successfully switched on Ferenterlogic for Fermenter {}".format(
                                        self.fermenter.id
                                    )
                                )
                        except Exception as e:
                            logging.error(
                                "Failed to switch on FermenterLogic {} {}".format(
                                    self.fermenter.id, e
                                )
                            )
                except:
                    pass
        except:
            pass


@parameters(
    [
        Property.Number(
            label="HeaterOffsetOn",
            configurable=True,
            description="Offset as decimal number when the heater is switched on. Should be greater then 'HeaterOffsetOff'. For example a value of 2 switches on the heater if the current temperature is 2 degrees below the target temperature",
        ),
        Property.Number(
            label="HeaterOffsetOff",
            configurable=True,
            description="Offset as decimal number when the heater is switched off. Should be smaller then 'HeaterOffsetOn'. For example a value of 1 switches off the heater if the current temperature is 1 degree below the target temperature",
        ),
        Property.Number(
            label="CoolerOffsetOn",
            configurable=True,
            description="Offset as decimal number when the cooler is switched on. Should be greater then 'CoolerOffsetOff'. For example a value of 2 switches on the cooler if the current temperature is 2 degrees above the target temperature",
        ),
        Property.Number(
            label="CoolerOffsetOff",
            configurable=True,
            description="Offset as decimal number when the cooler is switched off. Should be smaller then 'CoolerOffsetOn'. For example a value of 1 switches off the cooler if the current temperature is 1 degree above the target temperature",
        ),
        Property.Number(
            label="HeaterMaxPower",
            configurable=True,
            description="Max Power [%] for Heater (default: 100)",
        ),
        Property.Number(
            label="CoolerMaxPower",
            configurable=True,
            description="Max Power [%] for Cooler (default: 100)",
        ),
        Property.Select(
            label="AutoStart",
            options=["Yes", "No"],
            description="Autostart Fermenter on cbpi start",
        ),
        Property.Sensor(
            label="sensor2", description="Optional Sensor for LCDisplay(e.g. iSpindle)"
        ),
    ]
)
class FermenterHysteresis(CBPiFermenterLogic):

    async def run(self):
        try:
            self.heater_offset_min = float(self.props.get("HeaterOffsetOn", 0.2))
            self.heater_offset_max = float(self.props.get("HeaterOffsetOff", 0))
            self.cooler_offset_min = float(self.props.get("CoolerOffsetOn", 0.2))
            self.cooler_offset_max = float(self.props.get("CoolerOffsetOff", 0))
            self.heater_max_power = int(self.props.get("HeaterMaxPower", 100))
            self.cooler_max_power = int(self.props.get("CoolerMaxPower", 100))

            self.fermenter = self.get_fermenter(self.id)
            self.heater = self.fermenter.heater
            self.cooler = self.fermenter.cooler

            heater = self.cbpi.actor.find_by_id(self.heater)
            cooler = self.cbpi.actor.find_by_id(self.cooler)

            while self.running == True:

                sensor_value = float(
                    self.get_sensor_value(self.fermenter.sensor).get("value")
                )
                target_temp = float(self.get_fermenter_target_temp(self.id))

                try:
                    heater_state = heater.instance.state
                except:
                    heater_state = False
                try:
                    cooler_state = cooler.instance.state
                except:
                    cooler_state = False

                if sensor_value + self.heater_offset_min <= target_temp:
                    if self.heater and (heater_state == False):
                        await self.actor_on(self.heater, self.heater_max_power)

                if sensor_value + self.heater_offset_max >= target_temp:
                    if self.heater and (heater_state == True):
                        await self.actor_off(self.heater)

                if sensor_value >= self.cooler_offset_min + target_temp:
                    if self.cooler and (cooler_state == False):
                        await self.actor_on(self.cooler, self.cooler_max_power)

                if sensor_value <= self.cooler_offset_max + target_temp:
                    if self.cooler and (cooler_state == True):
                        await self.actor_off(self.cooler)

                await asyncio.sleep(1)

        except asyncio.CancelledError as e:
            pass
        except Exception as e:
            logging.error("Fermenter Hysteresis Error {}".format(e))
        finally:
            self.running = False
            if self.heater:
                await self.actor_off(self.heater)
            if self.cooler:
                await self.actor_off(self.cooler)


@parameters(
    [
        Property.Number(
            label="HeaterOffsetOn",
            configurable=True,
            description="Offset as decimal number when the heater is switched on. Should be greater then 'HeaterOffsetOff'. For example a value of 2 switches on the heater if the current temperature is 2 degrees below the target temperature",
        ),
        Property.Number(
            label="HeaterOffsetOff",
            configurable=True,
            description="Offset as decimal number when the heater is switched off. Should be smaller then 'HeaterOffsetOn'. For example a value of 1 switches off the heater if the current temperature is 1 degree below the target temperature",
        ),
        Property.Number(
            label="CoolerOffsetOn",
            configurable=True,
            description="Offset as decimal number when the cooler is switched on. Should be greater then 'CoolerOffsetOff'. For example a value of 2 switches on the cooler if the current temperature is 2 degrees below the target temperature",
        ),
        Property.Number(
            label="CoolerOffsetOff",
            configurable=True,
            description="Offset as decimal number when the cooler is switched off. Should be smaller then 'CoolerOffsetOn'. For example a value of 1 switches off the cooler if the current temperature is 1 degree below the target temperature",
        ),
        Property.Number(
            label="SpundingOffsetOpen",
            configurable=True,
            description="Offset above target pressure as decimal number when the valve is opened",
        ),
        Property.Select(
            label="ValveRelease",
            options=[1, 2, 3, 4, 5],
            description="Valve Release time in seconds",
        ),
        Property.Select(
            label="Pause",
            options=[1, 2, 3, 4, 5],
            description="Pause time in seconds between valve release",
        ),
        Property.Select(
            label="AutoStart",
            options=["Yes", "No"],
            description="Autostart Fermenter on cbpi start",
        ),
        Property.Sensor(
            label="sensor2", description="Optional Sensor for LCDisplay(e.g. iSpindle)"
        ),
    ]
)
class FermenterSpundingHysteresis(CBPiFermenterLogic):
    # subroutine that controls pressure
    async def pressure_control(self):
        self.spunding_offset = float(self.props.get("SpundingOffsetOpen", 0))
        self.valverelease = int(self.props.get("ValveRelease", 1))
        self.pause = int(self.props.get("Pause", 2))
        if self.valve and self.fermenter.pressure_sensor:
            # valve = self.cbpi.actor.find_by_id(self.valve)

            await self.actor_off(self.valve)
            # logging.info("Closing Spunding Valve")

            while self.running:
                target_pressure = float(self.fermenter.target_pressure)
                current_pressure = float(
                    self.get_sensor_value(self.fermenter.pressure_sensor).get("value")
                )
                # logging.info(f'Target: {target_pressure} | Current: {current_pressure}')
                if (
                    current_pressure >= (target_pressure + self.spunding_offset)
                    and target_pressure != 0
                ):
                    while current_pressure >= target_pressure:
                        await self.actor_on(self.valve)
                        await asyncio.sleep(self.valverelease)
                        await self.actor_off(self.valve)
                        await asyncio.sleep(self.pause)
                        current_pressure = float(
                            self.get_sensor_value(self.fermenter.pressure_sensor).get(
                                "value"
                            )
                        )
                        # logging.info("Value higher than target: Spunding loop is running")

                await asyncio.sleep(1)
        else:
            logging.info("No valve or pressure sensor defined")

    async def temperature_control(self):
        self.heater_offset_min = float(self.props.get("HeaterOffsetOn", 0.2))
        self.heater_offset_max = float(self.props.get("HeaterOffsetOff", 0))
        self.cooler_offset_min = float(self.props.get("CoolerOffsetOn", 0.2))
        self.cooler_offset_max = float(self.props.get("CoolerOffsetOff", 0))

        heater = self.cbpi.actor.find_by_id(self.heater)
        cooler = self.cbpi.actor.find_by_id(self.cooler)

        while self.running == True:

            sensor_value = float(
                self.get_sensor_value(self.fermenter.sensor).get("value")
            )
            target_temp = float(self.get_fermenter_target_temp(self.id))

            try:
                heater_state = heater.instance.state
            except:
                heater_state = False
            try:
                cooler_state = cooler.instance.state
            except:
                cooler_state = False

            if sensor_value + self.heater_offset_min <= target_temp:
                if self.heater and (heater_state == False):
                    await self.actor_on(self.heater)

            if sensor_value + self.heater_offset_max >= target_temp:
                if self.heater and (heater_state == True):
                    await self.actor_off(self.heater)

            if sensor_value >= self.cooler_offset_min + target_temp:
                if self.cooler and (cooler_state == False):
                    await self.actor_on(self.cooler)

            if sensor_value <= self.cooler_offset_max + target_temp:
                if self.cooler and (cooler_state == True):
                    await self.actor_off(self.cooler)

            await asyncio.sleep(1)

    async def run(self):
        try:
            self.fermenter = self.get_fermenter(self.id)
            self.heater = self.fermenter.heater
            self.cooler = self.fermenter.cooler
            self.valve = self.fermenter.valve

            pressure_controller = asyncio.create_task(self.pressure_control())
            temperature_controller = asyncio.create_task(self.temperature_control())

            await pressure_controller
            await temperature_controller

        except asyncio.CancelledError as e:
            pass
        except Exception as e:
            logging.error("Fermenter Spunding Hysteresis Error {}".format(e))
        finally:
            self.running = False
            if self.heater:
                await self.actor_off(self.heater)
            if self.cooler:
                await self.actor_off(self.cooler)
            if self.valve:
                await self.actor_off(self.valve)


def setup(cbpi):
    """
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    """
    cbpi.plugin.register("Fermenter Spunding Hysteresis", FermenterSpundingHysteresis)
    cbpi.plugin.register("Fermenter Hysteresis", FermenterHysteresis)
    cbpi.plugin.register("Fermenter Autostart", FermenterAutostart)



import json
import logging
import os.path
import re
from os import listdir
from os.path import isfile, join

import shortuuid
import yaml

from ..api.step import StepMove, StepResult, StepState


class RecipeController:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.logger = logging.getLogger(__name__)

    def urlify(self, s):

        # Remove all non-word characters (everything except numbers and letters)
        s = re.sub(r"[^\w\s]", "", s)

        # Replace all runs of whitespace with a single dash
        s = re.sub(r"\s+", "-", s)

        return s

    async def create(self, name):
        id = shortuuid.uuid()
        path = self.cbpi.config_folder.get_recipe_file_by_id(id)
        data = dict(
            basic=dict(name=name, author=self.cbpi.config.get("AUTHOR", "John Doe")),
            steps=[],
        )
        with open(path, "w") as file:
            yaml.dump(data, file)
        return id

    async def save(self, name, data):
        path = self.cbpi.config_folder.get_recipe_file_by_id(name)
        logging.info(data)
        with open(path, "w") as file:
            yaml.dump(data, file, indent=4, sort_keys=True)

    async def get_recipes(self):
        path = self.cbpi.config_folder.get_file_path("recipes")
        onlyfiles = [
            os.path.splitext(f)[0]
            for f in listdir(path)
            if isfile(join(path, f)) and f.endswith(".yaml")
        ]

        result = []
        for filename in onlyfiles:
            recipe_path = self.cbpi.config_folder.get_recipe_file_by_id(filename)
            with open(recipe_path) as file:
                data = yaml.load(file, Loader=yaml.FullLoader)
                dataset = data["basic"]
                dataset["file"] = filename
                result.append(dataset)
        return result

    async def get_by_name(self, name):

        recipe_path = self.cbpi.config_folder.get_recipe_file_by_id(name)
        with open(recipe_path) as file:
            return yaml.load(file, Loader=yaml.FullLoader)

    async def remove(self, name):
        path = self.cbpi.config_folder.get_recipe_file_by_id(name)
        os.remove(path)

    async def brew(self, name):

        recipe_path = self.cbpi.config_folder.get_recipe_file_by_id(name)
        with open(recipe_path) as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
            await self.cbpi.step.load_recipe(data)

    async def clone(self, id, new_name):
        recipe_path = self.cbpi.config_folder.get_recipe_file_by_id(id)
        with open(recipe_path) as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
            data["basic"]["name"] = new_name
            new_id = shortuuid.uuid()
            await self.save(new_id, data)

            return new_id



import logging
import os
import pathlib
import re

from aiohttp import streamer, web
from cbpi import __codename__, __version__
from cbpi.api import request_mapping
from cbpi.controller.system_controller import SystemController
from cbpi.job.aiohttp import get_scheduler_from_app
from cbpi.utils import json_dumps


class SystemHttpEndpoints:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.controller: SystemController = cbpi.system
        self.cbpi.register(self, url_prefix="/system")

    @request_mapping("/", method="GET", auth_required=False)
    async def state(self, request):
        """
        ---
        description: Get complete system state
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        plugin_list = await self.cbpi.plugin.load_plugin_list("cbpi4gui")
        try:
            version = plugin_list[0].get("Version", "not detected")
        except:
            version = "not detected"

        spindle = self.cbpi.config.get("spindledata", "No")
        if spindle == "Yes":
            spindledata = True
        else:
            spindledata = False

        return web.json_response(
            data=dict(
                actor=self.cbpi.actor.get_state(),
                fermenter=self.cbpi.fermenter.get_state(),
                sensor=self.cbpi.sensor.get_state(),
                kettle=self.cbpi.kettle.get_state(),
                step=self.cbpi.step.get_state(),
                fermentersteps=self.cbpi.fermenter.get_fermenter_steps(),
                config=self.cbpi.config.get_state(),
                notifications=self.cbpi.notification.get_state(),
                bf_recipes=await self.cbpi.upload.get_brewfather_recipes(0),
                version=__version__,
                spindledata=spindledata,
                guiversion=version,
                codename=__codename__,
            ),
            dumps=json_dumps,
        )

    @request_mapping(path="/logs", auth_required=False)
    async def http_get_log(self, request):
        result = []
        file_pattern = re.compile("^(\w+.).log(.?\d*)")
        for filename in sorted(os.listdir(self.cbpi.logsFolderPath), reverse=True):
            if file_pattern.match(filename):
                result.append(filename)
        return web.json_response(result)

    @request_mapping(path="/logs/{name}", method="DELETE", auth_required=False)
    async def delete_log(self, request):
        log_name = request.match_info["name"]
        self.cbpi.log.delete_log(log_name)

    @request_mapping(path="/logs", method="DELETE", auth_required=False)
    async def delete_all_logs(self, request):
        self.cbpi.log.delete_logs()
        return web.Response(status=204)

    @request_mapping(
        "/events", method="GET", name="get_all_events", auth_required=False
    )
    def get_all_events(self, request):
        """
        ---
        description: Get list of all registered events
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        return web.json_response(data=self.cbpi.bus.dump())

    @request_mapping("/jobs", method="GET", name="get_jobs", auth_required=False)
    def get_all_jobs(self, request):
        """
        ---
        description: Get all running Jobs
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        scheduler = get_scheduler_from_app(self.cbpi.app)
        result = []
        for j in scheduler:
            try:
                result.append(dict(name=j.name, type=j.type, time=j.start_time))
            except:
                pass
        return web.json_response(data=result)

    @request_mapping(
        "/restart", method="POST", name="RestartServer", auth_required=False
    )
    async def restart(self, request):
        """
        ---
        description: Restart System
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        await self.controller.restart()
        return web.Response(text="RESTART")

    @request_mapping(
        "/shutdown", method="POST", name="ShutdownSerer", auth_required=False
    )
    async def shutdown(self, request):
        """
        ---
        description: Shutdown System
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        await self.controller.shutdown()
        return web.Response(text="SHUTDOWN")

    @request_mapping("/backup", method="GET", name="BackupConfig", auth_required=False)
    async def backup(self, request):
        """
        ---
        description: Zip and download Config Folder
        tags:
        - System
        responses:
            "200":
                description: successful operation
                content:  # Response body
                application/zip:  # Media type
        """
        filename = await self.controller.backupConfig()
        # filename = "cbpi4_config.zip"
        file_name = pathlib.Path(os.path.join(".", filename))

        response = web.StreamResponse(
            status=200,
            reason="OK",
            headers={"Content-Type": "application/zip"},
        )
        await response.prepare(request)
        with open(file_name, "rb") as file:
            for line in file.readlines():
                await response.write(line)

        await response.write_eof()
        return response

    @request_mapping(
        "/log/{logtime}/", method="GET", name="BackupConfig", auth_required=False
    )
    async def downloadlog(self, request):
        """
        ---
        description: Zip and download craftbeerpi.service log
        tags:
        - System
        parameters:
        - name: "logtime"
          in: "path"
          description: "Logtime in hours"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "200":
                description: successful operation
                content:  # Response body
                application/zip:  # Media type
        """
        checklogtime = False
        logtime = request.match_info["logtime"]

        try:
            test = int(logtime)
            checklogtime = True
        except:
            if logtime == "b":
                checklogtime = True

        if checklogtime:
            await self.controller.downloadlog(logtime)
            filename = "cbpi4_log.zip"
            file_name = pathlib.Path(os.path.join(".", filename))

            response = web.StreamResponse(
                status=200,
                reason="OK",
                headers={"Content-Type": "application/zip"},
            )
            await response.prepare(request)
            with open(file_name, "rb") as file:
                for line in file.readlines():
                    await response.write(line)

            await response.write_eof()
            os.remove(file_name)
            return response
        else:
            return web.Response(status=400, text='Need integer or "b" for logtime.')

    @request_mapping(
        "/restore", method="POST", name="RestoreConfig", auth_required=False
    )
    async def restore(self, request):
        """
        ---
        description: Restore Config
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        logging.info("Restore Config")
        data = await request.post()
        logging.info("Data received")
        await self.controller.restoreConfig(data)
        return web.Response(status=200)

    @request_mapping(
        "/systeminfo", method="GET", name="SystemInfo", auth_required=False
    )
    async def systeminfo(self, request):
        """
        ---
        description: System Information
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        systeminfo = await self.controller.systeminfo()
        return web.json_response(data=systeminfo)

    @request_mapping("/uploadsvg", method="POST", name="UploadSVG", auth_required=False)
    async def uploadSVG(self, request):
        """
        ---
        description: Upload SVG file to widgets folder
        tags:
        - System
        responses:
            "200":
                description: successful operation
        """
        logging.info("Upload SVG file")
        data = await request.post()
        logging.info("Data received")
        await self.controller.uploadSVG(data)
        return web.Response(status=200)



import asyncio
import logging
from unittest.mock import MagicMock, patch

from cbpi.api import *

logger = logging.getLogger(__name__)

try:
    import RPi.GPIO as GPIO
except Exception:
    logger.warning("Failed to load RPi.GPIO. Using Mock instead")
    MockRPi = MagicMock()
    modules = {"RPi": MockRPi, "RPi.GPIO": MockRPi.GPIO}
    patcher = patch.dict("sys.modules", modules)
    patcher.start()
    import RPi.GPIO as GPIO

mode = GPIO.getmode()
if mode == None:
    GPIO.setmode(GPIO.BCM)


@parameters(
    [
        Property.Select(
            label="GPIO",
            options=[
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                20,
                21,
                22,
                23,
                24,
                25,
                26,
                27,
            ],
        ),
        Property.Select(
            label="Inverted",
            options=["Yes", "No"],
            description="No: Active on high; Yes: Active on low",
        ),
        Property.Select(
            label="SamplingTime",
            options=[2, 5],
            description="Time in seconds for power base interval (Default:5)",
        ),
    ]
)
class GPIOActor(CBPiActor):

    # Custom property which can be configured by the user
    @action(
        "Set Power",
        parameters=[
            Property.Number(
                label="Power", configurable=True, description="Power Setting [0-100]"
            )
        ],
    )
    async def setpower(self, Power=100, **kwargs):
        self.power = int(Power)
        if self.power < 0:
            self.power = 0
        if self.power > 100:
            self.power = 100
        await self.set_power(self.power)

    def get_GPIO_state(self, state):
        # ON
        if state == 1:
            return 1 if self.inverted == False else 0
        # OFF
        if state == 0:
            return 0 if self.inverted == False else 1

    async def on_start(self):
        self.power = None
        self.gpio = self.props.GPIO
        self.inverted = True if self.props.get("Inverted", "No") == "Yes" else False
        self.sampleTime = int(self.props.get("SamplingTime", 5))
        GPIO.setup(self.gpio, GPIO.OUT)
        GPIO.output(self.gpio, self.get_GPIO_state(0))
        self.state = False

    async def on(self, power=None):
        if power is not None:
            self.power = power
        else:
            self.power = 100
        #        await self.set_power(self.power)

        logger.info("ACTOR %s ON - GPIO %s " % (self.id, self.gpio))
        GPIO.output(self.gpio, self.get_GPIO_state(1))
        self.state = True

    async def off(self):
        logger.info("ACTOR %s OFF - GPIO %s " % (self.id, self.gpio))
        GPIO.output(self.gpio, self.get_GPIO_state(0))
        self.state = False

    def get_state(self):
        return self.state

    async def run(self):
        while self.running == True:
            if self.state == True:
                heating_time = self.sampleTime * (self.power / 100)
                wait_time = self.sampleTime - heating_time
                if heating_time > 0:
                    # logging.info("Heating Time: {}".format(heating_time))
                    GPIO.output(self.gpio, self.get_GPIO_state(1))
                    await asyncio.sleep(heating_time)
                if wait_time > 0:
                    # logging.info("Wait Time: {}".format(wait_time))
                    GPIO.output(self.gpio, self.get_GPIO_state(0))
                    await asyncio.sleep(wait_time)
            else:
                await asyncio.sleep(1)

    async def set_power(self, power):
        self.power = power
        await self.cbpi.actor.actor_update(self.id, power)
        pass


@parameters(
    [
        Property.Select(
            label="GPIO",
            options=[
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                20,
                21,
                22,
                23,
                24,
                25,
                26,
                27,
            ],
        ),
        Property.Number(label="Frequency", configurable=True),
        Property.Select(
            label="Inverted",
            options=["Yes", "No"],
            description="Inverts PWM load if set to yes (e.g. 90% = 10%). Default: No",
        ),
    ]
)
class GPIOPWMActor(CBPiActor):

    # Custom property which can be configured by the user
    @action(
        "Set Power",
        parameters=[
            Property.Number(
                label="Power", configurable=True, description="Power Setting [0-100]"
            )
        ],
    )
    async def setpower(self, Power=100, **kwargs):
        logging.info(Power)
        self.power = int(Power)
        if self.power < 0:
            self.power = 0
        if self.power > 100:
            self.power = 100
        await self.set_power(self.power)

    async def on_start(self):
        self.gpio = self.props.get("GPIO", None)
        self.inverted = self.props.get("Inverted", "No")
        self.frequency = self.props.get("Frequency", 0.5)
        if self.gpio is not None:
            GPIO.setup(self.gpio, GPIO.OUT)
            if self.inverted == "No":
                GPIO.output(self.gpio, 0)
            else:
                GPIO.output(self.gpio, 1)
        self.state = False
        self.power = None
        self.p = None
        pass

    async def on(self, power=None):
        logging.debug("PWM Actor Power: {}".format(power))
        if power is not None:
            self.power = power
        else:
            self.power = 100

        logging.debug("PWM Final Power: {}".format(self.power))

        logger.debug(
            "PWM ACTOR %s ON - GPIO %s - Frequency %s - Power %s"
            % (self.id, self.gpio, self.frequency, self.power)
        )
        try:
            if self.p is None:
                self.p = GPIO.PWM(int(self.gpio), float(self.frequency))
            if self.inverted == "No":
                self.p.start(self.power)
            else:
                self.p.start(100 - self.power)
            self.state = True
        #            await self.cbpi.actor.actor_update(self.id,self.power)
        except:
            pass

    async def off(self):
        logger.info("PWM ACTOR %s OFF - GPIO %s " % (self.id, self.gpio))
        if self.inverted == "No":
            self.p.ChangeDutyCycle(0)
        else:
            self.p.ChangeDutyCycle(100)

        self.state = False

    async def set_power(self, power):
        if self.p and self.state == True:
            if self.inverted == "No":
                self.p.ChangeDutyCycle(power)
            else:
                self.p.ChangeDutyCycle(
                    100 - power
                )  # Set power to 100-value to invert output
        await self.cbpi.actor.actor_update(self.id, power)
        pass

    def get_state(self):
        return self.state

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)


def setup(cbpi):
    """
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    """

    cbpi.plugin.register("GPIOActor", GPIOActor)
    cbpi.plugin.register("GPIOPWMActor", GPIOPWMActor)



import asyncio
import json
import logging
from contextlib import AsyncExitStack, asynccontextmanager
from re import M

import shortuuid
from aiomqtt import Client, MqttError, Will
from cbpi import __version__


class SatelliteController:

    def __init__(self, cbpi):
        self.client_id = shortuuid.uuid()
        self.cbpi = cbpi
        self.kettlecontroller = cbpi.kettle
        self.fermentercontroller = cbpi.fermenter
        self.sensorcontroller = cbpi.sensor
        self.actorcontroller = cbpi.actor
        self.logger = logging.getLogger(__name__)
        self.host = cbpi.static_config.get("mqtt_host", "localhost")
        self.port = cbpi.static_config.get("mqtt_port", 1883)
        self.username = cbpi.static_config.get("mqtt_username", None)
        self.password = cbpi.static_config.get("mqtt_password", None)
        self.client = None
        self.topic_filters = [
            ("cbpi/actor/+/on", self._actor_on),
            ("cbpi/actor/+/off", self._actor_off),
            ("cbpi/actor/+/power", self._actor_power),
            ("cbpi/updateactor", self._actorupdate),
            ("cbpi/updatekettle", self._kettleupdate),
            ("cbpi/updatesensor", self._sensorupdate),
            ("cbpi/updatefermenter", self._fermenterupdate),
        ]
        self.tasks = set()

    def remove_key(self, d, key):
        r = dict(d)
        del r[key]
        return r

    async def init(self):

        self.client = Client(
            self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            will=Will(topic="cbpi/disconnect", payload="CBPi Server Disconnected"),
            identifier=self.client_id,
        )
        try:
            ## Listen for mqtt messages in an (unawaited) asyncio task
            task = asyncio.create_task(self.listen())
            ## Save a reference to the task so it doesn't get garbage collected
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)
            self.logger.info("MQTT Connected to {}:{}".format(self.host, self.port))
        except asyncio.CancelledError as e:
            self.logger.error("MQTT Connection failed: {}".format(e))


    async def listen(self):
        while True:
            try:
                async with self.client as client:
                    await client.subscribe("#")
                    async for message in client.messages:
                        for topic_filter in self.topic_filters:
                            topic = topic_filter[0]
                            method = topic_filter[1]
                            if message.topic.matches(topic):
                                await method(message)
            except asyncio.CancelledError:
                # Cancel
                self.logger.warning("MQTT Listening Cancelled")
                break
            except Exception as e:
                self.logger.error("MQTT General Exception: {}".format(e))
                break
            except MqttError as e:
                self.logger.error("MQTT Exception: {}".format(e))
            await asyncio.sleep(5)

    async def publish(self, topic, message, retain=False):
        if self.client is not None and self.client._connected:
            try:
                await self.client.publish(topic, message, qos=1, retain=retain)
            except Exception as e:
                self.logger.warning("Failed to push data via mqtt: {}".format(e))

    async def _actor_on(self, message):
        try:
            topic_key = str(message.topic).split("/")
            await self.cbpi.actor.on(topic_key[2])
            self.logger.warning("Processed actor {} on via mqtt".format(topic_key[2]))
        except Exception as e:
            self.logger.warning("Failed to process actor on via mqtt: {}".format(e))

    async def _actor_off(self, message):
        try:
            topic_key = str(message.topic).split("/")
            await self.cbpi.actor.off(topic_key[2])
            self.logger.warning("Processed actor {} off via mqtt".format(topic_key[2]))
        except Exception as e:
            self.logger.warning("Failed to process actor off via mqtt: {}".format(e))

    async def _actor_power(self, message):
        try:
            topic_key = str(message.topic).split("/")
            try:
                power = int(message.payload.decode())
                if power > 100:
                    power = 100
                if power < 0:
                    power = 0
                await self.cbpi.actor.set_power(topic_key[2], power)
                # await self.cbpi.actor.actor_update(topic_key[2],power)
            except:
                self.logger.warning(
                    "Failed to set actor power via mqtt. No valid power in message"
                )
        except:
            self.logger.warning("Failed to set actor power via mqtt")

    async def _kettleupdate(self, message):
        try:
            self.kettle = self.kettlecontroller.get_state()
            for item in self.kettle["data"]:
                self.cbpi.push_update(
                    "cbpi/{}/{}".format("kettleupdate", item["id"]), item
                )
        except Exception as e:
            self.logger.warning("Failed to send kettleupdate via mqtt: {}".format(e))

    async def _fermenterupdate(self, message):
        try:
            self.fermenter = self.fermentercontroller.get_state()
            for item in self.fermenter["data"]:
                item_new = self.remove_key(item, "steps")
                self.cbpi.push_update(
                    "cbpi/{}/{}".format("fermenterupdate", item["id"]), item_new
                )
        except Exception as e:
            self.logger.warning("Failed to send fermenterupdate via mqtt: {}".format(e))

    async def _actorupdate(self, message):
        try:
            self.actor = self.actorcontroller.get_state()
            for item in self.actor["data"]:
                self.cbpi.push_update(
                    "cbpi/{}/{}".format("actorupdate", item["id"]), item
                )
        except Exception as e:
            self.logger.warning("Failed to send actorupdate via mqtt: {}".format(e))

    async def _sensorupdate(self, message):
        try:
            self.sensor = self.sensorcontroller.get_state()
            for item in self.sensor["data"]:
                self.cbpi.push_update(
                    "cbpi/{}/{}".format("sensorupdate", item["id"]), item
                )
        except Exception as e:
            self.logger.warning("Failed to send sensorupdate via mqtt: {}".format(e))

    def subscribe(self, topic, method):
        self.topic_filters.append((topic, method))
        return True

    def unsubscribe(self, topic, method):
        self.topic_filters.remove((topic, method))
        return True



import json
import logging
import os
from pathlib import Path

from cbpi.api.config import ConfigType
from cbpi.api.dataclasses import Config
from cbpi.utils import load_config


class ConfigController:

    def __init__(self, cbpi):
        self.cache = {}
        self.logger = logging.getLogger(__name__)
        self.cbpi = cbpi
        self.cbpi.register(self)
        self.path = cbpi.config_folder.get_file_path("config.json")
        self.path_static = cbpi.config_folder.get_file_path("config.yaml")
        self.logger.info(
            "Config folder path : "
            + os.path.join(Path(self.cbpi.config_folder.configFolderPath).absolute())
        )

    def get_state(self):
        result = {}
        for key, value in self.cache.items():
            result[key] = value.to_dict()
        return result

    async def init(self):
        self.static = load_config(self.path_static)
        with open(self.path) as json_file:
            data = json.load(json_file)
            for key, value in data.items():
                self.cache[key] = Config(
                    name=value.get("name"),
                    value=value.get("value"),
                    description=value.get("description"),
                    type=ConfigType(value.get("type", "string")),
                    source=value.get("source", "craftbeerpi"),
                    options=value.get("options", None),
                )

    def get(self, name, default=None):
        self.logger.debug("GET CONFIG VALUE %s (default %s)" % (name, default))
        if (
            name in self.cache
            and self.cache[name].value is not None
            and self.cache[name].value != ""
        ):
            return self.cache[name].value
        else:
            return default

    async def set(self, name, value):
        if name in self.cache:

            self.cache[name].value = value

            data = {}
            for key, value in self.cache.items():
                data[key] = value.to_dict()
            with open(self.path, "w") as file:
                json.dump(data, file, indent=4, sort_keys=True)

    async def add(
        self,
        name,
        value,
        type: ConfigType,
        description,
        source="craftbeerpi",
        options=None,
    ):
        self.cache[name] = Config(name, value, description, type, source, options)
        data = {}
        for key, value in self.cache.items():
            data[key] = value.to_dict()
        with open(self.path, "w") as file:
            json.dump(data, file, indent=4, sort_keys=True)

    async def remove(self, name):
        data = {}
        self.testcache = {}
        success = False
        for key, value in self.cache.items():
            try:
                if key != name:
                    data[key] = value.to_dict()
                    self.testcache[key] = Config(
                        name=data[key].get("name"),
                        value=data[key].get("value"),
                        description=data[key].get("description"),
                        type=ConfigType(data[key].get("type", "string")),
                        options=data[key].get("options", None),
                        source=data[key].get("source", "craftbeerpi"),
                    )
                    success = True
            except Exception as e:
                print(e)
                success = False
        if success == True:
            with open(self.path, "w") as file:
                json.dump(data, file, indent=4, sort_keys=True)
            self.cache = self.testcache

    async def obsolete(self, remove=False):
        result = {}
        for key, value in self.cache.items():
            if value.source not in ("craftbeerpi", "steps", "hidden"):
                test = await self.cbpi.plugin.load_plugin_list(value.source)
                if test == []:
                    update = self.get(str(value.source) + "_update")
                    if update:
                        result[str(value.source) + "_update"] = {"value": update}
                        if remove:
                            await self.remove(str(value.source) + "_update")
                    if remove:
                        await self.remove(key)
                    result[key] = value.to_dict()
        return result



from aiohttp import web
from cbpi.api import request_mapping
from cbpi.utils import json_dumps


class NotificationHttpEndpoints:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.cbpi.register(self, url_prefix="/notification")

    @request_mapping(
        path="/{id}/action/{action_id}", method="POST", auth_required=False
    )
    async def action(self, request):
        """
        ---
        description: Update an actor
        tags:
        - Notification
        parameters:
        - name: "id"
          in: "path"
          description: "Notification Id"
          required: true
          type: "string"
        - name: "action_id"
          in: "path"
          description: "Action Id"
          required: true
          type: "string"

        responses:
            "200":
                description: successful operation
        """

        notification_id = request.match_info["id"]
        action_id = request.match_info["action_id"]
        # print(notification_id, action_id)
        self.cbpi.notification.notify_callback(notification_id, action_id)
        return web.Response(status=200)

    @request_mapping("/delete", method="POST", auth_required=False)
    async def restart(self, request):
        """
        ---
        description: DeleteNotifications
        tags:
        - Notification
        responses:
            "200":
                description: successful operation
        """
        self.cbpi.notification.delete_all_notifications()
        return web.Response(status=200)



from setuptools import setup, find_packages
from cbpi import __version__
import platform

# read the contents of your README file
from os import popen, path
import os

localsystem = platform.system()
board_reqs = []
raspberrypi=False
if localsystem == "Linux":
    command="cat /proc/cpuinfo | grep 'Raspberry'"
    model=popen(command).read()
    if len(model) != 0:
        raspberrypi=True
        if os.path.exists("/proc/device-tree/compatible"):
          with open("/proc/device-tree/compatible", "rb") as f:
            compat = f.read()
            # Pi 5
            if b"brcm,bcm2712" in compat:
              board_reqs = [
              "rpi-lgpio"
              ]
            # Pi 4 and Earlier
            elif (
              b"brcm,bcm2835" in compat
              or b"brcm,bcm2836" in compat
              or b"brcm,bcm2837" in compat
              or b"brcm,bcm2838" in compat
              or b"brcm,bcm2711" in compat
              ):
              board_reqs = ["RPi.GPIO"]  

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='cbpi4',
      version=__version__,
      description='CraftBeerPi4 Brewing Software',
      author='Manuel Fritsch / Alexander Vollkopf',
      author_email='avollkopf@l@web.de',
      url='https://github.com/PiBrewing/craftbeerpi4',
      license='GPLv3',
      project_urls={
	    'Documentation': 'https://openbrewing.gitbook.io/craftbeerpi4_support/'},
      packages=find_packages(),
      include_package_data=True,
      package_data={
        # If any package contains *.txt or *.rst files, include them:
      '': ['*.txt', '*.rst', '*.yaml'],
      'cbpi': ['*','*.txt', '*.rst', '*.yaml']},

      python_requires='>=3.9',
      long_description=long_description,
	    long_description_content_type='text/markdown',
      install_requires=[
          "typing-extensions>=4",
          "aiohttp==3.11.14",
          "aiohttp-auth==0.1.1",
          "aiohttp-route-decorator==0.1.4",
          "aiohttp-security==0.5.0",
          "aiohttp-session==2.12.1",
          "aiohttp-swagger==1.0.16",
          "aiojobs==1.3.0 ",
          "aiosqlite==0.17.0",
          "cryptography==44.0.2",
          "pyopenssl==25.0.0",
          "requests==2.32.3",
          "voluptuous==0.14.2",
          "pyfiglet==1.0.2",
          'click==8.1.7',
          'shortuuid==1.0.13',
          'tabulate==0.9.0',
          'aiomqtt==2.3.0',
          'inquirer==3.4.0',
          'colorama==0.4.6',
          'psutil==7.0.0',
          'cbpi4gui',
          'importlib_metadata',
          'numpy==2.2.4',
          'pandas==2.2.3'] + board_reqs + (
          ['systemd-python'] if localsystem == "Linux" else [] ),

        dependency_links=[
        'https://testpypi.python.org/pypi',
        
        ],
      entry_points = {
        "console_scripts": [
            "cbpi=cbpi.cli:main",
        ]
    }
)



# -*- coding: utf-8 -*-
import asyncio
import base64
import logging
import os
import random
from logging.handlers import RotatingFileHandler
from unittest.mock import MagicMock, patch

import urllib3
from cbpi.api import *
from cbpi.api.config import ConfigType

logger = logging.getLogger(__name__)


class SensorLogTargetCSV(CBPiExtension):

    def __init__(self, cbpi):  # called from cbpi on start
        self.cbpi = cbpi
        self.logfiles = self.cbpi.config.get("CSVLOGFILES", "Yes")
        if self.logfiles == "No":
            return  # never run()
        self._task = asyncio.create_task(self.run())  # one time run() only

    async def run(self):  # called by __init__ once on start if CSV is enabled
        self.listener_ID = self.cbpi.log.add_sensor_data_listener(self.log_data_to_CSV)
        logger.info("CSV sensor log target listener ID: {}".format(self.listener_ID))

    async def log_data_to_CSV(
        self, cbpi, id: str, value: str, formatted_time, name
    ):  # called by log_data() hook from the log file controller
        self.logfiles = self.cbpi.config.get("CSVLOGFILES", "Yes")
        if self.logfiles == "No":
            # We intentionally do not unsubscribe the listener here because then we had no way of resubscribing him without a restart of cbpi
            # as long as cbpi was STARTED with CSVLOGFILES set to Yes this function is still subscribed, so changes can be made on the fly.
            # but after initially enabling this logging target a restart is required.
            return
        if id not in self.cbpi.log.datalogger:
            max_bytes = int(self.cbpi.config.get("SENSOR_LOG_MAX_BYTES", 100000))
            backup_count = int(self.cbpi.config.get("SENSOR_LOG_BACKUP_COUNT", 3))

            data_logger = logging.getLogger("cbpi.sensor.%s" % id)
            data_logger.propagate = False
            data_logger.setLevel(logging.DEBUG)
            handler = RotatingFileHandler(
                os.path.join(self.cbpi.log.logsFolderPath, f"sensor_{id}.log"),
                maxBytes=max_bytes,
                backupCount=backup_count,
            )
            data_logger.addHandler(handler)
            self.cbpi.log.datalogger[id] = data_logger

        self.cbpi.log.datalogger[id].info("%s,%s" % (formatted_time, str(value)))


def setup(cbpi):
    cbpi.plugin.register("SensorLogTargetCSV", SensorLogTargetCSV)



import asyncio
import logging
import time
import warnings
from datetime import datetime
from socket import timeout
from typing import KeysView

import numpy as np
from cbpi.api import *
from cbpi.api import Property, action, parameters
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType
from cbpi.api.dataclasses import (
    Fermenter,
    Kettle,
    NotificationAction,
    NotificationType,
    Props,
)
from cbpi.api.step import CBPiFermentationStep, StepResult
from cbpi.api.timer import Timer
from voluptuous.schema_builder import message


@parameters(
    [
        Property.Text(
            label="Notification", configurable=True, description="Text for notification"
        ),
        Property.Select(
            label="AutoNext",
            options=["Yes", "No"],
            description="Automatically move to next step (Yes) or pause after Notification (No)",
        ),
    ]
)
class FermenterNotificationStep(CBPiFermentationStep):

    async def NextStep(self, **kwargs):
        await self.next(self.fermenter.id)
        return StepResult.DONE

    async def on_timer_done(self, timer):
        self.summary = self.props.get("Notification", "")

        if self.AutoNext == True:
            self.cbpi.notify(
                self.name, self.props.get("Notification", ""), NotificationType.INFO
            )
            if self.shutdown != True:
                await self.next(self.fermenter.id)
                return StepResult.DONE
        else:
            self.cbpi.notify(
                self.name,
                self.props.get("Notification", ""),
                NotificationType.INFO,
                action=[NotificationAction("Next Step", self.NextStep)],
            )
            await self.push_update()

    async def on_timer_update(self, timer, seconds):
        await self.push_update()

    async def on_start(self):
        self.shutdown = False
        self.summary = ""
        self.AutoNext = False if self.props.get("AutoNext", "No") == "No" else True
        if self.timer is None:
            self.timer = Timer(
                1, on_update=self.on_timer_update, on_done=self.on_timer_done
            )
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        await self.push_update()

    async def run(self):
        while self.running == True:
            await asyncio.sleep(1)
            if self.timer.is_running is not True:
                self.timer.start()
                self.timer.is_running = True

        return StepResult.DONE


@parameters(
    [
        Property.Number(label="Temp", configurable=True),
        Property.Sensor(label="Sensor"),
        Property.Text(
            label="Notification",
            configurable=True,
            description="Text for notification when Temp is reached",
        ),
        Property.Select(
            label="AutoMode",
            options=["Yes", "No"],
            description="Switch Fermenterlogic automatically on and off -> Yes",
        ),
    ]
)
class FermenterTargetTempStep(CBPiFermentationStep):

    async def NextStep(self, **kwargs):
        if self.shutdown != True:
            await self.next(self.fermenter.id)
            return StepResult.DONE

    async def on_timer_done(self, timer):
        self.summary = ""
        await self.push_update()
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(
            self.name,
            self.props.get(
                "Notification",
                "Target Temp reached. Please add malt and klick next to move on.",
            ),
        )
        if self.shutdown == False:
            await self.next(self.fermenter.id)
            return StepResult.DONE

    async def on_timer_update(self, timer, seconds):
        await self.push_update()

    async def on_start(self):
        self.shutdown = False
        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        if self.fermenter is not None:
            self.fermenter.target_temp = float(self.props.get("Temp", 0))
            self.fermenter.target_pressure = 0
        if self.AutoMode == True:
            await self.setAutoMode(True)
        self.summary = "Waiting for Target Temp"
        if self.fermenter is not None and self.timer is None:
            self.timer = Timer(
                1, on_update=self.on_timer_update, on_done=self.on_timer_done
            )
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def run(self):
        while self.get_sensor_value(self.props.get("Sensor", None)).get("value") > 900:
            await asyncio.sleep(1)
        self.starttemp = self.get_sensor_value(self.props.get("Sensor", None)).get(
            "value"
        )
        if self.fermenter.target_temp >= self.starttemp:
            logging.info("warmup")
            while self.running == True:
                sensor_value = self.get_sensor_value(
                    self.props.get("Sensor", None)
                ).get("value")
                if (
                    sensor_value >= self.fermenter.target_temp
                    and self.timer.is_running is not True
                ):
                    self.timer.start()
                    self.timer.is_running = True
                await asyncio.sleep(1)
        elif self.fermenter.target_temp <= self.starttemp:
            logging.info("Cooldown")
            while self.running == True:
                sensor_value = self.get_sensor_value(
                    self.props.get("Sensor", None)
                ).get("value")
                if (
                    sensor_value <= self.fermenter.target_temp
                    and self.timer.is_running is not True
                ):
                    self.timer.start()
                    self.timer.is_running = True
                await asyncio.sleep(1)
        await self.push_update()
        return StepResult.DONE

    async def reset(self):
        self.timer = Timer(
            1, on_update=self.on_timer_update, on_done=self.on_timer_done
        )
        self.timer.is_running == False

    async def setAutoMode(self, auto_state):
        try:
            if (
                self.fermenter.instance is None
                or self.fermenter.instance.state == False
            ) and (auto_state is True):
                await self.cbpi.fermenter.toggle(self.fermenter.id)
            elif (self.fermenter.instance.state == True) and (auto_state is False):
                await self.cbpi.fermenter.toggle(self.fermenter.id)
            await self.push_update()

        except Exception as e:
            logging.error(
                "Failed to switch on FermenterLogic {} {}".format(self.fermenter.id, e)
            )


@parameters(
    [
        Property.Number(label="TimerD", description="Timer Days", configurable=True),
        Property.Number(label="TimerH", description="Timer Hours", configurable=True),
        Property.Number(label="TimerM", description="Timer Minutes", configurable=True),
        Property.Number(
            label="Temp", configurable=True, description="Step Temperature"
        ),
        Property.Number(
            label="Pressure", configurable=True, description="Step Pressure"
        ),
        Property.Sensor(label="Sensor", description="Temperature Sensor"),
        Property.Select(
            label="AutoMode",
            options=["Yes", "No"],
            description="Switch Fermenterlogic automatically on and off -> Yes",
        ),
    ]
)
class FermenterStep(CBPiFermentationStep):

    @action("Start Timer", [])
    async def start_timer(self):
        if self.timer.is_running is not True:
            self.cbpi.notify(self.name, "Timer started", NotificationType.INFO)
            self.timer.start()
            self.timer.is_running = True
            self.endtime = time.time() + self.fermentationtime
            await self.update_endtime()
            estimated_completion_time = datetime.fromtimestamp(
                time.time() + self.fermentationtime
            )
            self.cbpi.notify(
                self.name,
                "Timer started. Estimated completion: {}".format(
                    estimated_completion_time.strftime("%d.%m, %H:%M")
                ),
                NotificationType.INFO,
            )
        else:
            self.cbpi.notify(
                self.name, "Timer is already running", NotificationType.WARNING
            )

    #    @action("Add 1 Day to Timer", [])
    #    async def add_timer(self):
    #        if self.timer.is_running == True:
    #            self.cbpi.notify(self.name, '1 Day added', NotificationType.INFO)
    #            await self.timer.add(86400)
    #            self.endtime = self.endtime +86400
    #            await self.update_endtime()
    #        else:
    #            self.cbpi.notify(self.name, 'Timer must be running to add time', NotificationType.WARNING)

    async def on_timer_done(self, timer):
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(self.name, "Step finished", NotificationType.SUCCESS)
        if self.shutdown == False:
            await self.next(self.fermenter.id)
            return StepResult.DONE

    async def on_timer_update(self, timer, seconds):
        self.summary = Timer.format_time(seconds)
        await self.push_update()

    async def on_start(self):
        self.shutdown = False
        if self.endtime == 0:
            timeD = int(self.props.get("TimerD", 0))
            timeH = int(self.props.get("TimerH", 0))
            timeM = int(self.props.get("TimerM", 0))
            self.fermentationtime = (timeM + (60 * timeH) + (1440 * timeD)) * 60
        else:
            self.fermentationtime = self.endtime - time.time()

        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        if self.fermenter is not None:
            self.fermenter.target_temp = float(self.props.get("Temp", 0))
            self.fermenter.target_pressure = float(self.props.get("Pressure", 0))
        if self.AutoMode == True:
            await self.setAutoMode(True)
        await self.push_update()

        if self.fermenter is not None and self.timer is None:
            logging.info("Set Timer")
            self.timer = Timer(
                self.fermentationtime,
                on_update=self.on_timer_update,
                on_done=self.on_timer_done,
            )
            self.timer.is_running = False
        elif self.fermenter is not None:
            try:
                if self.timer.is_running == True:
                    self.timer.start()
                    self.endtime = time.time() + self.fermentationtime
                    await self.update_endtime()
            except:
                pass

        if (
            self.endtime != 0
            and self.timer is not None
            and self.timer.is_running == False
        ):
            self.timer.start()
            self.timer.is_running = True
            estimated_completion_time = datetime.fromtimestamp(
                time.time() + self.fermentationtime
            )
            self.cbpi.notify(
                self.name,
                "Timer restarted. Estimated completion: {}".format(
                    estimated_completion_time.strftime("%d.%m, %H:%M")
                ),
                NotificationType.INFO,
            )

        self.summary = "Waiting for Target Temp"
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def reset(self):
        timeD = int(self.props.get("TimerD", 0))
        timeH = int(self.props.get("TimerH", 0))
        timeM = int(self.props.get("TimerM", 0))
        self.fermentationtime = (timeM + (60 * timeH) + (1440 * timeD)) * 60
        self.timer = Timer(
            self.fermentationtime,
            on_update=self.on_timer_update,
            on_done=self.on_timer_done,
        )
        self.endtime = 0
        self.timer.is_running == False

    async def run(self):
        while self.get_sensor_value(self.props.get("Sensor", None)).get("value") > 900:
            await asyncio.sleep(1)
        self.starttemp = self.get_sensor_value(self.props.get("Sensor", None)).get(
            "value"
        )

        if self.fermenter.target_temp >= self.starttemp:
            logging.info("warmup")
            while self.running == True:
                await asyncio.sleep(1)
                sensor_value = self.get_sensor_value(
                    self.props.get("Sensor", None)
                ).get("value")
                if (
                    sensor_value >= self.fermenter.target_temp
                    and self.timer.is_running is not True
                ):
                    self.timer.start()
                    self.timer.is_running = True
                    self.endtime = time.time() + self.fermentationtime
                    await self.update_endtime()
                    estimated_completion_time = datetime.fromtimestamp(
                        time.time() + self.fermentationtime
                    )
                    self.cbpi.notify(
                        self.name,
                        "Timer started. Estimated completion: {}".format(
                            estimated_completion_time.strftime("%d.%m, %H:%M")
                        ),
                        NotificationType.INFO,
                    )
        elif self.fermenter.target_temp <= self.starttemp:
            logging.info("cooldown")
            while self.running == True:
                await asyncio.sleep(1)
                sensor_value = self.get_sensor_value(
                    self.props.get("Sensor", None)
                ).get("value")
                if (
                    sensor_value <= self.fermenter.target_temp
                    and self.timer.is_running is not True
                ):
                    self.timer.start()
                    self.timer.is_running = True
                    self.endtime = time.time() + self.fermentationtime
                    await self.update_endtime()
                    estimated_completion_time = datetime.fromtimestamp(
                        time.time() + self.fermentationtime
                    )
                    self.cbpi.notify(
                        self.name,
                        "Timer started. Estimated completion: {}".format(
                            estimated_completion_time.strftime("%d.%m, %H:%M")
                        ),
                        NotificationType.INFO,
                    )

        return StepResult.DONE

    async def setAutoMode(self, auto_state):
        try:
            if (
                self.fermenter.instance is None
                or self.fermenter.instance.state == False
            ) and (auto_state is True):
                await self.cbpi.fermenter.toggle(self.fermenter.id)
            elif (self.fermenter.instance.state == True) and (auto_state is False):
                await self.cbpi.fermenter.toggle(self.fermenter.id)
            await self.push_update()

        except Exception as e:
            logging.error(
                "Failed to switch on FermenterLogic {} {}".format(self.fermenter.id, e)
            )


@parameters(
    [
        Property.Number(
            label="Temp", configurable=True, description="Ramp to this temp"
        ),
        Property.Number(
            label="Pressure", configurable=True, description="Step Pressure"
        ),
        Property.Number(
            label="RampRate",
            configurable=True,
            description="Ramp x °C/F per  day. Default: 1",
        ),
        Property.Sensor(label="Sensor", description="Temperature Sensor"),
        Property.Text(
            label="Notification",
            configurable=True,
            description="Text for notification when Temp is reached",
        ),
        Property.Select(
            label="AutoMode",
            options=["Yes", "No"],
            description="Switch Fermenterlogic automatically on and off -> Yes",
        ),
    ]
)
class FermenterRampTempStep(CBPiFermentationStep):

    async def NextStep(self, **kwargs):
        if self.shutdown == False:
            await self.next(self.fermenter.id)
            return StepResult.DONE

    async def on_timer_done(self, timer):
        self.summary = ""
        await self.push_update()
        if self.AutoMode == True:
            await self.setAutoMode(False)
        self.cbpi.notify(
            self.name,
            self.props.get(
                "Notification",
                "Target Temp reached. Please add malt and klick next to move on.",
            ),
        )
        await self.next(self.fermenter.id)
        return StepResult.DONE

    async def on_timer_update(self, timer, seconds):
        await self.push_update()

    async def on_start(self):
        self.shutdown = False
        self.AutoMode = True if self.props.get("AutoMode", "No") == "Yes" else False
        self.rate = float(self.props.get("RampRate", 1))
        logging.info(self.rate)
        self.target_temp = round(float(self.props.get("Temp", 0)) * 10) / 10
        logging.info(self.target_temp)
        self.fermenter.target_pressure = float(self.props.get("Pressure", 0))
        while self.get_sensor_value(self.props.get("Sensor", None)).get("value") > 900:
            await asyncio.sleep(1)
        self.starttemp = self.get_sensor_value(self.props.get("Sensor", None)).get(
            "value"
        )

        self.current_target_temp = self.starttemp
        if self.fermenter is not None:
            await self.set_fermenter_target_temp(
                self.fermenter.id, self.current_target_temp
            )
        if self.AutoMode == True:
            await self.setAutoMode(True)
        self.summary = "Ramping to {}° with {}° per day".format(
            self.target_temp, self.rate
        )
        if self.fermenter is not None and self.timer is None:
            self.timer = Timer(
                1, on_update=self.on_timer_update, on_done=self.on_timer_done
            )
        await self.push_update()

    async def on_stop(self):
        await self.timer.stop()
        self.summary = ""
        if self.AutoMode == True:
            await self.setAutoMode(False)
        await self.push_update()

    async def calc_target_temp(self):
        delta_time = time.time() - self.starttime
        current_target_temp = (
            round((self.starttemp + delta_time * self.ratesecond) * 10) / 10
        )
        #        logging.info(current_target_temp)
        if current_target_temp != self.current_target_temp:
            self.current_target_temp = current_target_temp
            await self.set_fermenter_target_temp(
                self.fermenter.id, self.current_target_temp
            )
            # self.fermenter.target_temp = self.current_target_temp
            await self.push_update()

        pass

    async def run(self):
        self.delta_temp = self.target_temp - self.starttemp
        try:
            self.deltadays = abs(self.delta_temp / self.rate)
            self.deltaseconds = self.deltadays * 24 * 60 * 60
            self.ratesecond = self.delta_temp / self.deltaseconds
        except Exception as e:
            logging.info(e)
        self.starttime = time.time()

        if self.target_temp >= self.starttemp:
            logging.info("warmup")
            while self.running == True:
                if self.current_target_temp != self.target_temp:
                    await self.calc_target_temp()
                sensor_value = self.get_sensor_value(
                    self.props.get("Sensor", None)
                ).get("value")
                if (
                    sensor_value >= self.target_temp
                    and self.timer.is_running is not True
                ):
                    self.timer.start()
                    self.timer.is_running = True
                await asyncio.sleep(1)
        elif self.target_temp <= self.starttemp:
            logging.info("Cooldown")
            while self.running == True:
                if self.current_target_temp != self.target_temp:
                    await self.calc_target_temp()
                sensor_value = self.get_sensor_value(
                    self.props.get("Sensor", None)
                ).get("value")
                if (
                    sensor_value <= self.target_temp
                    and self.timer.is_running is not True
                ):
                    self.timer.start()
                    self.timer.is_running = True
                await asyncio.sleep(1)
        await self.push_update()
        return StepResult.DONE

    async def reset(self):
        self.timer = Timer(
            1, on_update=self.on_timer_update, on_done=self.on_timer_done
        )
        self.timer.is_running == False

    async def setAutoMode(self, auto_state):
        try:
            if (
                self.fermenter.instance is None
                or self.fermenter.instance.state == False
            ) and (auto_state is True):
                await self.cbpi.fermenter.toggle(self.fermenter.id)
            elif (self.fermenter.instance.state == True) and (auto_state is False):
                await self.fermenter.instance.stop()
            await self.push_update()

        except Exception as e:
            logging.error(
                "Failed to switch on FermenterLogic {} {}".format(self.fermenter.id, e)
            )


def setup(cbpi):
    """
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    """

    cbpi.plugin.register("FermenterNotificationStep", FermenterNotificationStep)
    cbpi.plugin.register("FermenterTargetTempStep", FermenterTargetTempStep)
    cbpi.plugin.register("FermenterRampTempStep", FermenterRampTempStep)
    cbpi.plugin.register("FermenterStep", FermenterStep)



import asyncio
import glob
import json
import logging
import os
import random
import shutil
import threading
import time
from unittest.mock import MagicMock, patch

import yaml
from aiohttp import web
from cbpi import __version__
from cbpi.api import *
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType

logger = logging.getLogger(__name__)


class ConfigUpdate(CBPiExtension):

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self._task = asyncio.create_task(self.run())

    def append_to_yaml(self, file_path, data_to_append):

        with open(file_path[0], "a+") as file:
            file.seek(0)
            yaml.dump(data_to_append, file, default_flow_style=False)

    async def run(self):
        logging.info("Check Config for required changes")

        # check is default steps are config parameters

        TEMP_UNIT = self.cbpi.config.get("TEMP_UNIT", "C")
        default_boil_temp = 99 if TEMP_UNIT == "C" else 212
        default_cool_temp = 20 if TEMP_UNIT == "C" else 68
        boil_temp = self.cbpi.config.get("steps_boil_temp", None)
        cooldown_sensor = self.cbpi.config.get("steps_cooldown_sensor", None)
        cooldown_actor = self.cbpi.config.get("steps_cooldown_actor", None)
        cooldown_temp = self.cbpi.config.get("steps_cooldown_temp", None)
        mashin_step = self.cbpi.config.get("steps_mashin", None)
        mash_step = self.cbpi.config.get("steps_mash", None)
        mashout_step = self.cbpi.config.get("steps_mashout", None)
        boil_step = self.cbpi.config.get("steps_boil", None)
        cooldown_step = self.cbpi.config.get("steps_cooldown", None)
        max_dashboard_number = self.cbpi.config.get("max_dashboard_number", None)
        current_dashboard_number = self.cbpi.config.get(
            "current_dashboard_number", None
        )
        logfiles = self.cbpi.config.get("CSVLOGFILES", None)
        influxdb = self.cbpi.config.get("INFLUXDB", None)
        influxdbaddr = self.cbpi.config.get("INFLUXDBADDR", None)
        influxdbname = self.cbpi.config.get("INFLUXDBNAME", None)
        influxdbuser = self.cbpi.config.get("INFLUXDBUSER", None)
        influxdbpwd = self.cbpi.config.get("INFLUXDBPWD", None)
        influxdbcloud = self.cbpi.config.get("INFLUXDBCLOUD", None)
        influxdbport = self.cbpi.config.get("INFLUXDBPORT", None)
        influxdbmeasurement = self.cbpi.config.get("INFLUXDBMEASUREMENT", None)
        mqttupdate = self.cbpi.config.get("MQTTUpdate", None)
        PRESSURE_UNIT = self.cbpi.config.get("PRESSURE_UNIT", None)
        SENSOR_LOG_BACKUP_COUNT = self.cbpi.config.get("SENSOR_LOG_BACKUP_COUNT", None)
        SENSOR_LOG_MAX_BYTES = self.cbpi.config.get("SENSOR_LOG_MAX_BYTES", None)
        slow_pipe_animation = self.cbpi.config.get("slow_pipe_animation", None)
        NOTIFY_ON_ERROR = self.cbpi.config.get("NOTIFY_ON_ERROR", None)
        PLAY_BUZZER = self.cbpi.config.get("PLAY_BUZZER", None)
        BoilAutoTimer = self.cbpi.config.get("BoilAutoTimer", None)
        MASH_TUN = self.cbpi.config.get("MASH_TUN", None)
        AutoMode = self.cbpi.config.get("AutoMode", None)
        AddMashIn = self.cbpi.config.get("AddMashInStep", None)
        bfuserid = self.cbpi.config.get("brewfather_user_id", None)
        bfapikey = self.cbpi.config.get("brewfather_api_key", None)
        bflistlength = self.cbpi.config.get("brewfather_list_length", None)
        RecipeCreationPath = self.cbpi.config.get("RECIPE_CREATION_PATH", None)
        BoilKettle = self.cbpi.config.get("BoilKettle", None)
        CONFIG_STATUS = self.cbpi.config.get("CONFIG_STATUS", None)
        self.version = __version__
        current_grid = self.cbpi.config.get("current_grid", None)
        mqtt_offset = self.cbpi.static_config.get("mqtt_offset", None)

        if boil_temp is None:
            logger.info("INIT Boil Temp Setting")
            try:
                await self.cbpi.config.add(
                    "steps_boil_temp",
                    default_boil_temp,
                    type=ConfigType.NUMBER,
                    description="Default Boil Temperature for Recipe Creation",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_boil_temp",
                    boil_temp,
                    type=ConfigType.NUMBER,
                    description="Default Boil Temperature for Recipe Creation",
                    source="steps",
                )

        if cooldown_sensor is None:
            logger.info("INIT Cooldown Sensor Setting")
            try:
                await self.cbpi.config.add(
                    "steps_cooldown_sensor",
                    "",
                    type=ConfigType.SENSOR,
                    description="Alternative Sensor to monitor temperature durring cooldown (if not selected, Kettle Sensor will be used)",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_cooldown_sensor",
                    cooldown_sensor,
                    type=ConfigType.SENSOR,
                    description="Alternative Sensor to monitor temperature durring cooldown (if not selected, Kettle Sensor will be used)",
                    source="steps",
                )

        if cooldown_actor is None:
            logger.info("INIT Cooldown Actor Setting")
            try:
                await self.cbpi.config.add(
                    "steps_cooldown_actor",
                    "",
                    type=ConfigType.ACTOR,
                    description="Actor to trigger cooldown water on and off (default: None)",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_cooldown_actor",
                    cooldown_actor,
                    type=ConfigType.ACTOR,
                    description="Actor to trigger cooldown water on and off (default: None)",
                    source="steps",
                )

        if cooldown_temp is None:
            logger.info("INIT Cooldown Temp Setting")
            try:
                await self.cbpi.config.add(
                    "steps_cooldown_temp",
                    default_cool_temp,
                    type=ConfigType.NUMBER,
                    description="Cooldown temp will send notification when this temeprature is reached",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_cooldown_temp",
                    cooldown_temp,
                    type=ConfigType.NUMBER,
                    description="Cooldown temp will send notification when this temeprature is reached",
                    source="steps",
                )

        if cooldown_step is None:
            logger.info("INIT Cooldown Step Type")
            try:
                await self.cbpi.config.add(
                    "steps_cooldown",
                    "",
                    type=ConfigType.STEP,
                    description="Cooldown step type",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_cooldown",
                    cooldown_step,
                    type=ConfigType.STEP,
                    description="Cooldown step type",
                    source="steps",
                )

        if mashin_step is None:
            logger.info("INIT MashIn Step Type")
            try:
                await self.cbpi.config.add(
                    "steps_mashin",
                    "",
                    type=ConfigType.STEP,
                    description="MashIn step type",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_mashin",
                    mashin_step,
                    type=ConfigType.STEP,
                    description="MashIn step type",
                    source="steps",
                )

        if mash_step is None:
            logger.info("INIT Mash Step Type")
            try:
                await self.cbpi.config.add(
                    "steps_mash",
                    "",
                    type=ConfigType.STEP,
                    description="Mash step type",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_mash",
                    mash_step,
                    type=ConfigType.STEP,
                    description="Mash step type",
                    source="steps",
                )

        if mashout_step is None:
            logger.info("INIT MashOut Step Type")
            try:
                await self.cbpi.config.add(
                    "steps_mashout",
                    "",
                    type=ConfigType.STEP,
                    description="MashOut step type",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_mashout",
                    mashout_step,
                    type=ConfigType.STEP,
                    description="MashOut step type",
                    source="steps",
                )

        if boil_step is None:
            logger.info("INIT Boil Step Type")
            try:
                await self.cbpi.config.add(
                    "steps_boil",
                    "",
                    type=ConfigType.STEP,
                    description="Boil step type",
                    source="steps",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "steps_boil",
                    boil_step,
                    type=ConfigType.STEP,
                    description="Boil step type",
                    source="steps",
                )

        if max_dashboard_number is None:
            logger.info("INIT Max Dashboard Numbers for multiple dashboards")
            try:
                await self.cbpi.config.add(
                    "max_dashboard_number",
                    4,
                    type=ConfigType.SELECT,
                    description="Max Number of Dashboards",
                    source="craftbeerpi",
                    options=[
                        {"label": "1", "value": 1},
                        {"label": "2", "value": 2},
                        {"label": "3", "value": 3},
                        {"label": "4", "value": 4},
                        {"label": "5", "value": 5},
                        {"label": "6", "value": 6},
                        {"label": "7", "value": 7},
                        {"label": "8", "value": 8},
                        {"label": "9", "value": 9},
                        {"label": "10", "value": 10},
                    ],
                )

            except:
                logger.warning("Unable to update database")

        if current_dashboard_number is None:
            logger.info("INIT Current Dashboard Number")
            try:
                await self.cbpi.config.add(
                    "current_dashboard_number",
                    1,
                    type=ConfigType.NUMBER,
                    description="Number of current Dashboard",
                    source="hidden",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "current_dashboard_number",
                    current_dashboard_number,
                    type=ConfigType.NUMBER,
                    description="Number of current Dashboard",
                    source="hidden",
                )

        ## Check if AtuoMode for Steps is in config

        if AutoMode is None:
            logger.info("INIT AutoMode")
            try:
                await self.cbpi.config.add(
                    "AutoMode",
                    "Yes",
                    type=ConfigType.SELECT,
                    description="Use AutoMode in steps",
                    source="steps",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )

            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "AutoMode",
                    AutoMode,
                    type=ConfigType.SELECT,
                    description="Use AutoMode in steps",
                    source="steps",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )

        ## Check if AddMashInStep for Steps is in config

        if AddMashIn is None:
            logger.info("INIT AddMashInStep")
            try:
                await self.cbpi.config.add(
                    "AddMashInStep",
                    "Yes",
                    type=ConfigType.SELECT,
                    description="Add MashIn Step automatically if not defined in recipe",
                    source="steps",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )

            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "AddMashInStep",
                    AddMashIn,
                    type=ConfigType.SELECT,
                    description="Add MashIn Step automatically if not defined in recipe",
                    source="steps",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )

        ## Check if Brewfather UserID is in config

        if bfuserid is None:
            logger.info("INIT Brewfather User ID")
            try:
                await self.cbpi.config.add(
                    "brewfather_user_id",
                    "",
                    type=ConfigType.STRING,
                    description="Brewfather User ID",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")

        ## Check if Brewfather API Key is in config

        if bfapikey is None:
            logger.info("INIT Brewfather API Key")
            try:
                await self.cbpi.config.add(
                    "brewfather_api_key",
                    "",
                    type=ConfigType.STRING,
                    description="Brewfather API Key",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")

        ## Check if Brewfather API Key is in config

        if bflistlength is None:
            logger.info("INIT Brewfather Recipe List Length")
            try:
                await self.cbpi.config.add(
                    "brewfather_list_length",
                    50,
                    type=ConfigType.SELECT,
                    description="Brewfather Recipe List length",
                    source="craftbeerpi",
                    options=[
                        {"label": "5", "value": 5},
                        {"label": "10", "value": 10},
                        {"label": "25", "value": 25},
                        {"label": "50", "value": 50},
                        {"label": "100", "value": 100},
                    ],
                )
            except:
                logger.warning("Unable to update config")

        ## Check if Brewfather API Key is in config

        if RecipeCreationPath is None:
            logger.info("INIT Recipe Creation Path")
            try:
                await self.cbpi.config.add(
                    "RECIPE_CREATION_PATH",
                    "upload",
                    type=ConfigType.STRING,
                    description="API path to creation plugin. Default: upload . CHANGE ONLY IF USING A RECIPE CREATION PLUGIN",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")

        ## Check if  Kettle for Boil, Whirlpool and Cooldown is in config

        if BoilKettle is None:
            logger.info("INIT BoilKettle")
            try:
                await self.cbpi.config.add(
                    "BoilKettle",
                    "",
                    type=ConfigType.KETTLE,
                    description="Define Kettle that is used for Boil, Whirlpool and Cooldown. If not selected, MASH_TUN will be used",
                    source="steps",
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "BoilKettle",
                    BoilKettle,
                    type=ConfigType.KETTLE,
                    description="Define Kettle that is used for Boil, Whirlpool and Cooldown. If not selected, MASH_TUN will be used",
                    source="steps",
                )

        if MASH_TUN is None:
            logger.info("INIT MASH_TUN")
            try:
                await self.cbpi.config.add(
                    "MASH_TUN",
                    "",
                    type=ConfigType.KETTLE,
                    description="Default Mash Tun",
                    source="steps",
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "MASH_TUN",
                    MASH_TUN,
                    type=ConfigType.KETTLE,
                    description="Default Mash Tun",
                    source="steps",
                )

        ## Check if CSV logfiles is on config
        if logfiles is None:
            logger.info("INIT CSV logfiles")
            try:
                await self.cbpi.config.add(
                    "CSVLOGFILES",
                    "Yes",
                    type=ConfigType.SELECT,
                    description="Write sensor data to csv logfiles (enabling requires restart)",
                    source="craftbeerpi",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                try:
                    await self.cbpi.config.add(
                        "CSVLOGFILES",
                        logfiles,
                        type=ConfigType.SELECT,
                        description="Write sensor data to csv logfiles (enabling requires restart)",
                        source="craftbeerpi",
                        options=[
                            {"label": "Yes", "value": "Yes"},
                            {"label": "No", "value": "No"},
                        ],
                    )
                except:
                    logger.warning("Unable to update config")

        ## Check if influxdb is on config
        if influxdb is None:
            logger.info("INIT Influxdb")
            try:
                await self.cbpi.config.add(
                    "INFLUXDB",
                    "No",
                    type=ConfigType.SELECT,
                    description="Write sensor data to influxdb (enabling requires restart)",
                    source="craftbeerpi",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                try:
                    await self.cbpi.config.add(
                        "INFLUXDB",
                        influxdb,
                        type=ConfigType.SELECT,
                        description="Write sensor data to influxdb (enabling requires restart)",
                        source="craftbeerpi",
                        options=[
                            {"label": "Yes", "value": "Yes"},
                            {"label": "No", "value": "No"},
                        ],
                    )
                except:
                    logger.warning("Unable to update config")

        ## Check if influxdbport is in config and remove it as it is obsolete
        if influxdbport is not None:
            logger.warning("Remove obsolete Influxdbport config parameter")
            try:
                await self.cbpi.config.remove("INFLUXDBPORT")
            except:
                logger.warning("Unable to update config")

        ## Check if influxdbaddr is in config
        if influxdbaddr is None:
            logger.info("INIT Influxdbaddr")
            try:
                await self.cbpi.config.add(
                    "INFLUXDBADDR",
                    "http://localhost:8086",
                    type=ConfigType.STRING,
                    description="URL Address of your influxdb server incl. http:// and port, e.g. http://localhost:8086 (If INFLUXDBCLOUD set to Yes use URL Address of your influxdb cloud server)",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                try:
                    await self.cbpi.config.add(
                        "INFLUXDBADDR",
                        influxdbaddr,
                        type=ConfigType.STRING,
                        description="URL Address of your influxdb server incl. http:// and port, e.g. http://localhost:8086 (If INFLUXDBCLOUD set to Yes use URL Address of your influxdb cloud server)",
                        source="craftbeerpi",
                    )
                except:
                    logger.warning("Unable to update config")

        ## Check if influxdbname is in config
        if influxdbname is None:
            logger.info("INIT Influxdbname")
            try:
                await self.cbpi.config.add(
                    "INFLUXDBNAME",
                    "cbpi4",
                    type=ConfigType.STRING,
                    description="Name of your influxdb database name (If INFLUXDBCLOUD set to Yes use bucket of your influxdb cloud database)",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                try:
                    await self.cbpi.config.add(
                        "INFLUXDBNAME",
                        influxdbname,
                        type=ConfigType.STRING,
                        description="Name of your influxdb database name (If INFLUXDBCLOUD set to Yes use bucket of your influxdb cloud database)",
                        source="craftbeerpi",
                    )
                except:
                    logger.warning("Unable to update config")

        ## Check if influxduser is in config
        if influxdbuser is None:
            logger.info("INIT Influxdbuser")
            try:
                await self.cbpi.config.add(
                    "INFLUXDBUSER",
                    " ",
                    type=ConfigType.STRING,
                    description="User name for your influxdb database (only if required)(If INFLUXDBCLOUD set to Yes use organisation of your influxdb cloud database)",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                try:
                    await self.cbpi.config.add(
                        "INFLUXDBUSER",
                        influxdbuser,
                        type=ConfigType.STRING,
                        description="User Name for your influxdb database (only if required)(If INFLUXDBCLOUD set to Yes use organisation of your influxdb cloud database)",
                        source="craftbeerpi",
                    )
                except:
                    logger.warning("Unable to update config")

        ## Check if influxdpwd is in config
        if influxdbpwd is None:
            logger.info("INIT Influxdbpwd")
            try:
                await self.cbpi.config.add(
                    "INFLUXDBPWD",
                    " ",
                    type=ConfigType.STRING,
                    description="Password for your influxdb database (only if required)(If INFLUXDBCLOUD set to Yes use token of your influxdb cloud database)",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                try:
                    await self.cbpi.config.add(
                        "INFLUXDBPWD",
                        influxdbpwd,
                        type=ConfigType.STRING,
                        description="Password for your influxdb database (only if required)(If INFLUXDBCLOUD set to Yes use token of your influxdb cloud database)",
                        source="craftbeerpi",
                    )
                except:
                    logger.warning("Unable to update config")

        ## Check if influxdb cloud is on config
        if influxdbcloud is None:
            logger.info("INIT influxdbcloud")
            try:
                await self.cbpi.config.add(
                    "INFLUXDBCLOUD",
                    "No",
                    type=ConfigType.SELECT,
                    description="Write sensor data to influxdb cloud (INFLUXDB must set to Yes)",
                    source="craftbeerpi",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )
            except:
                logger.warning("Unable to update config")

                ## Check if influxdbname is in config
        if influxdbmeasurement is None:
            logger.info("INIT Influxdb measurementname")
            try:
                await self.cbpi.config.add(
                    "INFLUXDBMEASUREMENT",
                    "measurement",
                    type=ConfigType.STRING,
                    description="Name of the measurement in your INFLUXDB database (default: measurement)",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update config")

        if mqttupdate is None:
            logger.info("INIT MQTT update frequency for Kettles and Fermenters")
            try:
                await self.cbpi.config.add(
                    "MQTTUpdate",
                    0,
                    type=ConfigType.SELECT,
                    description="Forced MQTT Update frequency in s for Kettle and Fermenter (no changes in payload required). Restart required after change",
                    source="craftbeerpi",
                    options=[
                        {"label": "30", "value": 30},
                        {"label": "60", "value": 60},
                        {"label": "120", "value": 120},
                        {"label": "300", "value": 300},
                        {"label": "Never", "value": 0},
                    ],
                )
            except:
                logger.warning("Unable to update database")

        ## Check if PRESSURE_UNIT is in config
        if PRESSURE_UNIT is None:
            logger.info("INIT PRESSURE_UNIT")
            try:
                await self.cbpi.config.add(
                    "PRESSURE_UNIT",
                    "kPa",
                    type=ConfigType.SELECT,
                    description="Set unit for pressure",
                    source="craftbeerpi",
                    options=[
                        {"label": "kPa", "value": "kPa"},
                        {"label": "PSI", "value": "PSI"},
                    ],
                )
            except:
                logger.warning("Unable to update config")

        # check if SENSOR_LOG_BACKUP_COUNT exists in config
        if SENSOR_LOG_BACKUP_COUNT is None:
            logger.info("INIT SENSOR_LOG_BACKUP_COUNT")
            try:
                await self.cbpi.config.add(
                    "SENSOR_LOG_BACKUP_COUNT",
                    3,
                    type=ConfigType.NUMBER,
                    description="Max. number of backup logs",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update database")

        # check if SENSOR_LOG_MAX_BYTES exists in config
        if SENSOR_LOG_MAX_BYTES is None:
            logger.info("Init maximum size of sensor logfiles")
            try:
                await self.cbpi.config.add(
                    "SENSOR_LOG_MAX_BYTES",
                    100000,
                    type=ConfigType.NUMBER,
                    description="Max. number of bytes in sensor logs",
                    source="craftbeerpi",
                )
            except:
                logger.warning("Unable to update database")

        # Check if slow_pipe_animation is in config
        if slow_pipe_animation is None:
            logger.info("INIT slow_pipe_animation")
            try:
                await self.cbpi.config.add(
                    "slow_pipe_animation",
                    "Yes",
                    type=ConfigType.SELECT,
                    description="Slow down dashboard pipe animation taking up close to 100% of the CPU's capacity",
                    source="craftbeerpi",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )
            except:
                logger.warning("Unable to update config")

        ## Check if NOTIFY_ON_ERROR is in config
        if NOTIFY_ON_ERROR is None:
            logger.info("INIT NOTIFY_ON_ERROR")
            try:
                await self.cbpi.config.add(
                    "NOTIFY_ON_ERROR",
                    "No",
                    type=ConfigType.SELECT,
                    description="Send Notification on Logging Error",
                    source="craftbeerpi",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )
            except:
                logger.warning("Unable to update config")

        ## Check if PLAY_BUZZER is in config
        if PLAY_BUZZER is None:
            logger.info("INIT PLAY_BUZZER")
            try:
                await self.cbpi.config.add(
                    "PLAY_BUZZER",
                    "No",
                    type=ConfigType.SELECT,
                    description="Play buzzer sound in Web interface on Notifications",
                    source="craftbeerpi",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )
            except:
                logger.warning("Unable to update config")

        if BoilAutoTimer is None:
            logging.info("INIT BoilAutoTimer")
            try:
                await self.cbpi.config.add(
                    "BoilAutoTimer",
                    "No",
                    type=ConfigType.SELECT,
                    description="Start Boil timer automatically if Temp does not change for 5 Minutes and is above 95C/203F",
                    source="steps",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )

                BoilAutoTimer = self.cbpi.config.get("BoilAutoTimer", "No")
            except:
                logging.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "BoilAutoTimer",
                    BoilAutoTimer,
                    type=ConfigType.SELECT,
                    description="Start Boil timer automatically if Temp does not change for 5 Minutes and is above 95C/203F",
                    source="steps",
                    options=[
                        {"label": "Yes", "value": "Yes"},
                        {"label": "No", "value": "No"},
                    ],
                )

        if current_grid is None:
            logger.info("INIT Current Dashboard Number")
            try:
                await self.cbpi.config.add(
                    "current_grid",
                    5,
                    type=ConfigType.NUMBER,
                    description="Dashboard Grid Width",
                    source="hidden",
                )
            except:
                logger.warning("Unable to update database")
        else:
            if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
                await self.cbpi.config.add(
                    "current_grid",
                    current_grid,
                    type=ConfigType.NUMBER,
                    description="Dashboard Grid Width",
                    source="hidden",
                )

        # Check if CustomSVG props are correct for latest functionality (Widget dependent on acxtor state -> UI >= 0.3.14.a8)
        plugin_list = await self.cbpi.plugin.load_plugin_list("cbpi4gui")
        try:
            version = plugin_list[0].get("Version", "not detected")
        except:
            version = "not detected"

        logging.info(f"GUI Version: {version}")
        try:
            dashboard_files = glob.glob(
                self.cbpi.config_folder.get_dashboard_path("cbpi_dashboard*.json")
            )
            write = False

            for dashboard_file in dashboard_files:
                with open(dashboard_file, "r") as file:
                    data = json.load(file)

                for id in data["elements"]:
                    if id["type"] == "CustomSVG":
                        try:
                            testoff = id["props"]["WidgetOff"]
                        except:
                            id["props"]["WidgetOff"] = id["props"]["name"]
                            write = True
                        try:
                            teston = id["props"]["WidgetOn"]
                        except:
                            try:
                                id["props"]["WidgetOn"] = id["props"]["nameon"]
                                write = True
                            except:
                                pass

                if write:
                    with open(dashboard_file, "w") as outfile:
                        json.dump(data, outfile, indent=4, sort_keys=True)
        except Exception as e:
            logging.error(e)

        if mqtt_offset is None:
            logging.info("INIT MQTT Offset in static config")
            try:
                static_config_file = glob.glob(
                    self.cbpi.config_folder.get_file_path("config.yaml")
                )
                data_to_append = {"mqtt_offset": False}
                self.append_to_yaml(static_config_file, data_to_append)
                pass
            except Exception as e:
                logging.error(e)
                logging.warning("Unable to update database")

        ## Check if influxdbname is in config
        if CONFIG_STATUS is None or CONFIG_STATUS != self.version:
            logger.warning("Setting Config Status")
            try:
                await self.cbpi.config.add(
                    "CONFIG_STATUS",
                    self.version,
                    type=ConfigType.STRING,
                    description="Status of the config file. Internal use for maintenance",
                    source="hidden",
                )
            except:
                logger.warning("Unable to update config")


def setup(cbpi):
    cbpi.plugin.register("ConfigUpdate", ConfigUpdate)
    pass



# -*- coding: utf-8 -*-
import asyncio
import base64
import logging
import os
import random
from unittest.mock import MagicMock, patch

import urllib3
from cbpi.api import *
from cbpi.api.config import ConfigType
from urllib3 import PoolManager, Retry, Timeout

logger = logging.getLogger(__name__)


class SensorLogTargetInfluxDB(CBPiExtension):

    def __init__(self, cbpi):  # called from cbpi on start
        self.cbpi = cbpi
        self.influxdb = self.cbpi.config.get("INFLUXDB", "No")
        if self.influxdb == "No":
            return  # never run()
        self._task = asyncio.create_task(self.run())  # one time run() only
        self.counter = 0
        self.max_retries = 2
        self.send = True

    async def run(self):  # called by __init__ once on start if influx is enabled
        self.listener_ID = self.cbpi.log.add_sensor_data_listener(
            self.log_data_to_InfluxDB
        )
        logger.info(
            "InfluxDB sensor log target listener ID: {}".format(self.listener_ID)
        )

    async def log_data_to_InfluxDB(
        self, cbpi, id: str, value: str, timestamp, name
    ):  # called by log_data() hook from the log file controller
        self.influxdb = self.cbpi.config.get("INFLUXDB", "No")
        if self.influxdb == "No":
            # We intentionally do not unsubscribe the listener here because then we had no way of resubscribing him without a restart of cbpi
            # as long as cbpi was STARTED with INFLUXDB set to Yes this function is still subscribed, so changes can be made on the fly.
            # but after initially enabling this logging target a restart is required.
            return
        self.influxdbcloud = self.cbpi.config.get("INFLUXDBCLOUD", "No")
        self.influxdbaddr = self.cbpi.config.get("INFLUXDBADDR", None)
        self.influxdbname = self.cbpi.config.get("INFLUXDBNAME", None)
        self.influxdbuser = self.cbpi.config.get("INFLUXDBUSER", None)
        self.influxdbpwd = self.cbpi.config.get("INFLUXDBPWD", None)
        self.influxdbmeasurement = self.cbpi.config.get(
            "INFLUXDBMEASUREMENT", "measurement"
        )
        timeout = Timeout(connect=2.0, read=None)
        try:
            sensor = self.cbpi.sensor.find_by_id(id)
            if sensor is not None:
                itemname = sensor.name.replace(" ", "_")
                out = (
                    str(self.influxdbmeasurement)
                    + ",source="
                    + itemname
                    + ",itemID="
                    + str(id)
                    + " value="
                    + str(value)
                )
        except Exception as e:
            logging.error("InfluxDB ID Error: {}".format(e))

        if self.influxdbcloud == "Yes" and self.send == True:
            if self.counter <= self.max_retries:
                self.influxdburl = (
                    self.influxdbaddr
                    + "/api/v2/write?org="
                    + self.influxdbuser
                    + "&bucket="
                    + self.influxdbname
                    + "&precision=s"
                )
                try:
                    header = {
                        "User-Agent": id,
                        "Authorization": "Token {}".format(self.influxdbpwd),
                    }
                    http = PoolManager(timeout=timeout)
                    req = http.request(
                        "POST",
                        self.influxdburl,
                        body=out.encode(),
                        headers=header,
                        retries=Retry(2),
                    )
                    if req.status != 204:
                        raise Exception(f"InfluxDB Status code {req.status}")
                except Exception as e:
                    self.counter += 1
                    logging.error(
                        "InfluxDB cloud write Error #{}: {}".format(self.counter, e)
                    )
            else:
                logging.warning("Waiting 3 Minutes before connecting to INFLUXDB again")
                self.send = False
                await asyncio.sleep(180)
                self.counter = 0
                self.send = True

        elif self.influxdbcloud == "No" and self.send == True:
            if self.counter <= self.max_retries:
                self.base64string = base64.b64encode(
                    ("%s:%s" % (self.influxdbuser, self.influxdbpwd)).encode()
                )
                self.influxdburl = self.influxdbaddr + "/write?db=" + self.influxdbname
                try:
                    header = {
                        "User-Agent": id,
                        "Content-Type": "application/x-www-form-urlencoded",
                        "Authorization": "Basic %s" % self.base64string.decode("utf-8"),
                    }
                    http = PoolManager(timeout=timeout)
                    req = http.request(
                        "POST",
                        self.influxdburl,
                        body=out.encode(),
                        headers=header,
                        retries=Retry(2),
                    )
                    if req.status != 204:
                        raise Exception(f"InfluxDB Status code {req.status}")
                except Exception as e:
                    self.counter += 1
                    logging.error(
                        "InfluxDB write Error #{}: {}".format(self.counter, e)
                    )
            else:
                logging.warning("Waiting 3 Minutes before connecting to INFLUXDB again")
                self.send = False
                await asyncio.sleep(180)
                self.counter = 0
                self.send = True


def setup(cbpi):
    cbpi.plugin.register("SensorLogTargetInfluxDB", SensorLogTargetInfluxDB)



import logging

from cbpi.api.dataclasses import Actor
from cbpi.controller.basic_controller2 import BasicController
from tabulate import tabulate


class ActorController(BasicController):

    def __init__(self, cbpi):
        super(ActorController, self).__init__(cbpi, Actor, "actor.json")
        self.update_key = "actorupdate"
        self.sorting = True

    async def on(self, id, power=None):
        try:
            item = self.find_by_id(id)
            if power is None:
                logging.info("Power is none")
                if item.power:
                    power = item.power
                else:
                    power = 100
            if item.instance.state is False:
                await item.instance.on(power)
                # await self.push_udpate()
                self.cbpi.ws.send(
                    dict(
                        topic=self.update_key,
                        data=list(map(lambda item: item.to_dict(), self.data)),
                    ),
                    self.sorting,
                )
                self.cbpi.push_update(
                    "cbpi/actorupdate/{}".format(id), item.to_dict(), True
                )
            else:
                await self.set_power(id, power)

        except Exception as e:
            logging.error("Failed to switch on Actor {} {}".format(id, e))

    async def off(self, id):
        try:
            item = self.find_by_id(id)
            if item.instance.state is True:
                await item.instance.off()
                # await self.push_udpate()
                self.cbpi.ws.send(
                    dict(
                        topic=self.update_key,
                        data=list(map(lambda item: item.to_dict(), self.data)),
                    ),
                    self.sorting,
                )
                self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to switch on Actor {} {}".format(id, e), True)

    async def toogle(self, id):
        try:
            item = self.find_by_id(id)
            instance = item.get("instance")
            await instance.toggle()
            self.cbpi.ws.send(
                dict(
                    topic=self.update_key,
                    data=list(map(lambda item: item.to_dict(), self.data)),
                ),
                self.sorting,
            )
            self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to toggle Actor {} {}".format(id, e))

    async def set_power(self, id, power):
        try:
            item = self.find_by_id(id)
            await item.instance.set_power(power)
        except Exception as e:
            logging.error("Failed to set power {} {}".format(id, e))

    async def actor_update(self, id, power):
        try:
            item = self.find_by_id(id)
            item.power = round(power)
            # await self.push_udpate()
            self.cbpi.ws.send(
                dict(
                    topic=self.update_key,
                    data=list(map(lambda item: item.to_dict(), self.data)),
                ),
                self.sorting,
            )
            self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to update Actor {} {}".format(id, e))

    async def timeractor_update(self, id, timer):
        try:
            item = self.find_by_id(id)
            item.timer = round(timer)
            # await self.push_udpate()
            self.cbpi.ws.send(
                dict(
                    topic=self.update_key,
                    data=list(map(lambda item: item.to_dict(), self.data)),
                )
            )
            self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to update Actor {} {}".format(id, e))

    async def ws_actor_update(self):
        try:
            # await self.push_udpate()
            self.cbpi.ws.send(
                dict(
                    topic=self.update_key,
                    data=list(map(lambda x: x.to_dict(), self.data)),
                ),
                self.sorting,
            )
        #            self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to update Actors {}".format(e))



import asyncio
import logging
import os
from datetime import datetime
from email import message

import shortuuid
from cbpi.api import *
from cbpi.api.dataclasses import NotificationType


class NotificationController:

    def __init__(self, cbpi):
        """
        :param cbpi: craftbeerpi object
        """
        self.cbpi = cbpi
        self.logger = logging.getLogger(__name__)
        logging.root.addFilter(self.notify_log_event)
        self.callback_cache = {}
        self.listener = {}
        self.notifications = []
        self.update_key = "notificationupdate"
        self.sorting = False
        self.check_startup_message()

    def check_startup_message(self):
        self.restore_error = self.cbpi.config_folder.get_file_path("restore_error.log")
        try:
            with open(self.restore_error) as f:
                for line in f:
                    self.notifications.insert(
                        0,
                        [
                            f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}: Restore Error | {line}'
                        ],
                    )
            os.remove(self.restore_error)
        except Exception as e:
            pass

    def notify_log_event(self, record):
        NOTIFY_ON_ERROR = self.cbpi.config.get("NOTIFY_ON_ERROR", "No")
        if NOTIFY_ON_ERROR == "Yes":
            try:
                message = str(record.msg)
            except:
                message = record.msg
            try:
                if record.levelno > 20:
                    # on log events higher then INFO we want to notify all clients
                    type = NotificationType.WARNING
                    if record.levelno > 30:
                        type = NotificationType.ERROR
                    self.cbpi.notify(
                        title=f"{record.levelname}", message=message, type=type
                    )
            except Exception as e:
                pass
            finally:
                return True
        return True

    def get_state(self):
        result = self.notifications
        return result

    def add_listener(self, method):
        listener_id = shortuuid.uuid()
        self.listener[listener_id] = method
        return listener_id

    def remove_listener(self, listener_id):
        try:
            del self.listener[listener_id]
        except:
            self.logger.error("Failed to remove listener {}".format(listener_id))

    async def _call_listener(self, title, message, type, action):
        background_tasks = set()
        for id, method in self.listener.items():
            task = asyncio.create_task(method(self.cbpi, title, message, type, action))
            background_tasks.add(task)
            task.add_done_callback(background_tasks.discard)

    def notify(
        self,
        title,
        message: str,
        type: NotificationType = NotificationType.INFO,
        action=[],
        timeout: int = 5000,
    ) -> None:
        """
        This is a convinience method to send notification to the client

        :param key: notification key
        :param message: notification message
        :param type: notification type (info,warning,danger,successs)
        :return:
        """
        notifcation_id = shortuuid.uuid()
        background_tasks = set()

        def prepare_action(item):
            item.id = shortuuid.uuid()
            return item.to_dict()

        actions = list(map(lambda item: prepare_action(item), action))
        self.callback_cache[notifcation_id] = action
        self.cbpi.ws.send(
            dict(
                id=notifcation_id,
                topic="notifiaction",
                type=type.value,
                title=title,
                message=message,
                action=actions,
                timeout=timeout,
            )
        )
        data = dict(
            type=type.value,
            title=title,
            message=message,
            action=actions,
            timeout=timeout,
        )
        self.cbpi.push_update(topic="cbpi/notification", data=data)
        self.notifications.insert(
            0, [f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}: {title} | {message}']
        )
        if len(self.notifications) > 100:
            self.notifications = self.notifications[:100]
        self.cbpi.ws.send(
            dict(topic=self.update_key, data=self.notifications), self.sorting
        )
        task = asyncio.create_task(self._call_listener(title, message, type, action))
        background_tasks.add(task)
        task.add_done_callback(background_tasks.discard)

    def delete_all_notifications(self):
        self.notifications = []
        self.cbpi.ws.send(
            dict(topic=self.update_key, data=self.notifications), self.sorting
        )

    def notify_callback(self, notification_id, action_id) -> None:
        try:
            action = next(
                (
                    item
                    for item in self.callback_cache[notification_id]
                    if item.id == action_id
                ),
                None,
            )
            if action.method is not None:
                background_tasks = set()
                task = asyncio.create_task(action.method())
                background_tasks.add(task)
                task.add_done_callback(background_tasks.discard)
            del self.callback_cache[notification_id]
        except Exception as e:
            self.logger.error("Failed to call notification callback")
